import numpy as np
import os.path
from itertools import *
import pandas as pd
import logging
import time
import pysnptools.util as pstutil
from pysnptools.pstreader import PstReader
from pysnptools.kernelstandardizer import DiagKtoN


class KernelReader(PstReader):
    """A KernelReader is one of three things:

    * A class such as :class:`KernelNpz` for you to a file with data. For example,

        >>> from pysnptools.kernelreader import KernelNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> kernel_file = example_file('pysnptools/examples/toydata.kernel.npz')
        >>> kernel_on_disk = KernelNpz(kernel_file)
        >>> print(kernel_on_disk) # prints specification for reading from file
        KernelNpz('...pysnptools/examples/toydata.kernel.npz')
        >>> kernel_on_disk.iid_count # prints the number of iids (but doesn't read any kernel values)
        500

    * A :class:`.KernelData` class that holds kernel data in memory, typically after computing from a SnpReader or reading it from a KernelReader:

        >>> # Compute kernel from a SnpReader
        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.standardizer import Unit
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> bed_file = example_file('tests/datasets/all_chr.maf0.001.N300.*','*.bed')
        >>> snp_on_disk = Bed(bed_file,count_A1=False)
        >>> kerneldata1 = snp_on_disk.read_kernel(Unit()) #reads the SNP values and computes the kernel
        >>> type(kerneldata1.val).__name__ # The val property is an ndarray of kernel values
        'ndarray'
        >>> print(kerneldata1) # prints the specification of the in-memory kernel information
        KernelData(SnpKernel(Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False),standardizer=Unit()))
        >>> kerneldata1.iid_count #prints the number of iids (number of individuals) in this in-memory data
        300
        >>> # Read kernel from a KernelReader
        >>> kernel_file = example_file('pysnptools/examples/toydata.kernel.npz')
        >>> kernel_on_disk = KernelNpz(kernel_file)
        >>> kerneldata2 = kernel_on_disk.read() #reads the kernel values
        >>> print(kerneldata2) # prints the specification of the in-memory kernel information
        KernelData(KernelNpz(...pysnptools/examples/toydata.kernel.npz'))
        >>> kerneldata2.iid_count #prints the number of iids (number of individuals) in this in-memory data
        500


    * A subset of any KernelReader, specified with "[ *iid_index* ]" (or  specified with "[ *iid0_index* , *iid1_index* ]"), to read only some kernel values. It can
      also be used to re-order the values.

        >>> kernel_on_disk = KernelNpz(kernel_file)
        >>> subset_on_disk1 = kernel_on_disk[[3,4]] # specification for a subset of the data on disk. No kernel values are read yet.
        >>> print(subset_on_disk1.iid_count) # prints the number of iids in this subset (but still doesn't read any kernel values)
        2
        >>> print(subset_on_disk1) #prints a specification of 'subset_on_disk1'
        KernelNpz(...pysnptools/examples/toydata.kernel.npz')[[3,4],[3,4]]
        >>> kerneldata_subset = subset_on_disk1.read() # efficiently (if possible) reads the specified subset of values from the disk
        >>> print(kerneldata_subset) # prints the specification of the in-memory kernel information
        KernelData(KernelNpz(...pysnptools/examples/toydata.kernel.npz')[[3,4],[3,4]])
        >>> print((int(kerneldata_subset.val.shape[0]), int(kerneldata_subset.val.shape[1]))) # The dimensions of the ndarray of kernel values
        (2, 2)
        >>> subset_on_disk2 = kernel_on_disk[[3,4],::2] # specification for a subset of the data on disk. No kernel values are read yet.
        >>> print((subset_on_disk2.iid0_count, subset_on_disk2.iid1_count))
        (2, 250)


    The KernelReaders Classes

        ================================== ================== ====================== ================== ====================
        *Class*                            *Format*           *Random Access*        *Suffixes*         *Write* method?
        :class:`.KernelData`               in-memory          Yes                    *n/a*              *n/a*
        :class:`.KernelNpz`                binary             No                     .kernel.npz        Yes
        :class:`.KernelHdf5`               binary             Yes                    .kernel.hdf5       Yes
        :class:`.Identity`                 *n/a*              Yes                    *n/a*              No
        :class:`.SnpKernel`                depends            depends                *n/a*              No
        ================================== ================== ====================== ================== ====================


    Methods & Properties:

        Every KernelReader, such as :class:`.KernelNpz` and :class:`.KernelData`, when square has these properties: :attr:`iid`, :attr:`iid_count`,
        and these methods: :meth:`read`, and :meth:`iid_to_index`. A square kernel is one that has the same iid list for both its rows and columns.

        More generally, KernelReaders can have one iid list for its rows and a different iid list for its columns, so these properties and methods are also defined: :attr:`iid0`, :attr:`iid1`, :attr:`iid0_count`,
        :attr:`iid1_count`, :meth:`iid0_to_index`, and :meth:`iid1_to_index`.

        See below for details.

        :class:`.KernelData` is a KernelReader so it supports the above properties and methods. In addition, it supports property :attr:`KernelData.val`, method :meth:`.KernelData.standardize`, and equality testing.

        See below for details.

        Some of the classes, such as :class:`.KernelNpz`, also provide a static :meth:`KernelNpz.write` method for writing :class:`.KernelData`.

        >>> # create a kernel from a Bed file and write to KernelNpz format
        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> from pysnptools.standardizer import Unit
        >>> import pysnptools.util as pstutil

        >>> bed_file = example_file('pysnptools/examples/toydata.5chrom.bed')
        >>> kerneldata = Bed(bed_file,count_A1=False).read_kernel(Unit())     # Create a kernel from the data in the Bed file
        >>> pstutil.create_directory_if_necessary("tempdir/toydata.kernel.npz")
        >>> KernelNpz.write("tempdir/toydata.kernel.npz",kerneldata)      # Write data in KernelNpz format
        KernelNpz('tempdir/toydata.kernel.npz')

    iids:

        Individual are identified with an iid, which is a ndarray of two strings: a family ID and an individual ID. For example:

        >>> kernel_on_disk = KernelNpz(kernel_file)
        >>> print(kernel_on_disk.iid[:3]) # print the first three iids
        [['per0' 'per0']
         ['per1' 'per1']
         ['per2' 'per2']]
        >>> print(kernel_on_disk.iid_to_index([['per2','per2'],['per1','per1']])) #Find the indexes for two iids.
        [2 1]

    :class:`.KernelReader` is a kind of :class:`.PstReader`. See the documentation for :class:`.PstReader` to learn about:

        * When Data is Read
        * When Data is Re-Read and Copied
        * Avoiding Unwanted ndarray Allocations
        * Creating Subsetting PstReaders with Indexing

    The :meth:`read` Method

        By default the :meth:`read` returns a ndarray of numpy.float64 laid out in memory in F-contiguous order (iid0-index varies the fastest). You may, instead,
        ask for numpy.float32 or for C-contiguous order or any order. See :meth:`read` for details.

    The :meth:`.KernelData.standardize` Method
        The :meth:`.KernelData.standardize` method, available only on :class:`.KernelData`, does in-place standardization of the in-memory
        kernel data. The method multiples the values with a scalar factor such that the diagonal sums to iid_count. Although it works in place, for convenience
        it also returns itself. See :meth:`.KernelData.standardize` for details.

            >>> kernel_on_disk = KernelNpz(kernel_file)
            >>> kerneldata1 = kernel_on_disk.read() # read all kernel values into memory
            >>> print(np.diag(kerneldata1.val).sum())
            5000000.0
            >>> kerneldata1.standardize() # standardize changes the values in kerneldata1.val
            KernelData(KernelNpz(...pysnptools/examples/toydata.kernel.npz'))
            >>> print(np.diag(kerneldata1.val).sum())
            500.0
            >>> kerneldata2 = kernel_on_disk.read().standardize() # Read and standardize in one expression with only one ndarray allocated.
            >>> print(np.diag(kerneldata2.val).sum())
            500.0

    Details of Methods & Properties:
    """

    def __init__(self, *args, **kwargs):
        super(KernelReader, self).__init__(*args, **kwargs)

    @property
    def iid(self):
        """A ndarray of the iids. Each iid is a ndarray of two strings (a family ID and a individual ID) that identifies an individual.
        Assumes the kernel is square, so will throw an exception if the row iids are different from the column iids.

        :rtype: ndarray (length :attr:`.iid_count`) of ndarray (length 2) of strings

        This property (to the degree practical) reads only iid and sid data from the disk, not kernel value data. Moreover, the iid data is read from file only once.

        :Example:

        >>> from pysnptools.kernelreader import KernelNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> kernel_file = example_file('pysnptools/examples/toydata.kernel.npz')
        >>> kernel_on_disk = KernelNpz(kernel_file)
        >>> print(kernel_on_disk.iid[:3]) # print the first three iids
        [['per0' 'per0']
         ['per1' 'per1']
         ['per2' 'per2']]
        """
        assert (
            self.iid0 is self.iid1
        ), "When 'iid' is used, iid0 must be the same as iid1"
        return self.iid0

    @property
    def iid0(self):
        """
        A ndarray of the row iids. See :attr:`.iid`
        """
        return self.row

    @property
    def iid1(self):
        """
        A ndarray of the column iids. See :attr:`.iid`
        """
        return self.col

    @property
    def iid_count(self):
        """number of iids
        Assumes the kernel is square, so will throw an exception if the row iids are different from the column iids.

        :rtype: integer

        This property (to the degree practical) reads only iid data from the disk, not kernel value data. Moreover, the iid data is read from file only once.
        """
        assert (
            self.iid0 is self.iid1
        ), "When 'iid_count' is used, iid0 must be the same as iid1"
        return self.iid0_count

    @property
    def iid0_count(self):
        """number of row iids. See :attr:`iid_count`

        :rtype: integer
        """
        return self.row_count

    @property
    def iid1_count(self):
        """number of column iids. See :attr:`iid_count`

        :rtype: integer
        """
        return self.col_count

    @property
    def row_property(self):
        """
        Defined for compatibility with :class:`.PstReader`. Will always be empty.
        """
        if not hasattr(self, "_row_property"):
            self._row_property = np.empty((self.row_count, 0))
        return self._row_property

    @property
    def col_property(self):
        """
        Defined for compatibility with :class:`.PstReader`. Will always be empty.
        """
        if not hasattr(self, "_col_property"):
            self._col_property = np.empty((self.col_count, 0))
        return self._col_property

    # !!check that views always return contiguous memory by default
    def read(
        self,
        order="F",
        dtype=np.float64,
        force_python_only=False,
        view_ok=False,
        num_threads=None,
    ):
        """Reads the kernel values and returns a :class:`.KernelData` (with :attr:`KernelData.val` property containing a new ndarray of the kernel values).

        :param order: {'F' (default), 'C', 'A'}, optional -- Specify the order of the ndarray. If order is 'F' (default),
            then the array will be in F-contiguous order (iid0-index varies the fastest).
            If order is 'C', then the returned array will be in C-contiguous order (iid1-index varies the fastest).
            If order is 'A', then the :attr:`KernelData.val`
            ndarray may be in any order (either C-, Fortran-contiguous, or even discontiguous).
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`KernelData.val` ndarray.
        :type dtype: data-type

        :param force_python_only: optional -- If False (default), may use outside library code. If True, requests that the read
            be done without outside library code.
        :type force_python_only: bool

        :param view_ok: optional -- If False (default), allocates new memory for the :attr:`KernelData.val`'s ndarray. If True,
            if practical and reading from a :class:`KernelData`, will return a new
            :class:`KernelData` with a ndarray shares memory with the original :class:`KernelData`.
            Typically, you'll also wish to use "order='A'" to increase the chance that sharing will be possible.
            Use these parameters with care because any change to either ndarray (for example, via :meth:`.KernelData.standardize`) will effect
            the others. Also keep in mind that :meth:`read` relies on ndarray's mechanisms to decide whether to actually
            share memory and so it may ignore your suggestion and allocate a new ndarray anyway.
        :type view_ok: bool

        :param num_threads: optional -- The number of threads with which to standardize data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int

        :rtype: :class:`.KernelData`

        Calling the method again causes the kernel values to be re-read and creates a new in-memory :class:`.KernelData` with a new ndarray of kernel values.

        If you request the values for only a subset of the sids or iids, (to the degree practical) only that subset will be read from disk.

        :Example:

        >>> from pysnptools.kernelreader import KernelNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> npz_file = example_file('pysnptools/examples/toydata.kernel.npz')
        >>> kernel_on_disk = KernelNpz(npz_file)
        >>> kerneldata1 = kernel_on_disk.read() # Read all the kernel data returning a KernelData instance
        >>> print(type(kerneldata1.val).__name__) # The KernelData instance contains a ndarray of the data.
        ndarray
        >>> subset_kerneldata = kernel_on_disk[::2].read() # From the disk, read kernel values for every other iid
        >>> print('{0:.6f}'.format(subset_kerneldata.val[0,0])) # Print the first kernel value in the subset
        9923.069928
        >>> subsub_kerneldata = subset_kerneldata[:10].read(order='A',view_ok=True) # Create an in-memory subset of the subset with kernel values for the first ten iids. Share memory if practical.
        >>> import numpy as np
        >>> #print(np.may_share_memory(subset_kerneldata.val, subsub_kerneldata.val)) # Do the two ndarray's share memory? They could. Currently they won't.
        """
        dtype = np.dtype(dtype)
        val = self._read(
            None, None, order, dtype, force_python_only, view_ok, num_threads
        )
        from pysnptools.kernelreader import KernelData

        ret = KernelData(iid0=self.iid0, iid1=self.iid1, val=val, name=str(self))
        return ret

    def iid_to_index(self, list):
        """Takes a list of iids and returns a list of index numbers.
        Assumes the kernel is square, so will throw an exception if the row iids are different from the column iids.

        :param list: list of iids
        :type order: list of list of strings

        :rtype: ndarray of int

        This method (to the degree practical) reads only iid from the disk, not kernel value data. Moreover, the iid data is read from file only once.

        :Example:

        >>> from pysnptools.kernelreader import KernelNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> npz_file = example_file('pysnptools/examples/toydata.kernel.npz')
        >>> kernel_on_disk = KernelNpz(npz_file)
        >>> print(kernel_on_disk.iid_to_index([['per2','per2'],['per1','per1']])) #Find the indexes for two iids.
        [2 1]
        """
        assert (
            self.iid0 is self.iid1
        ), "When 'iid_to_index' is used, iid0 must be the same as iid1"
        return self.iid0_to_index(list)

    def iid0_to_index(self, list):
        """Takes a list of row iids and returns a list of index numbers. See :attr:`iid_to_index`"""
        return self.row_to_index(list)

    @staticmethod
    def _makekey(item):
        return tuple(str(i) for i in item)

    def iid1_to_index(self, list):
        """Takes a list of column iids and returns a list of index numbers. See :attr:`iid_to_index`"""
        return self.col_to_index(list)

    def __getitem__(self, iid_indexer_and_snp_indexer):
        from pysnptools.kernelreader._subset import _KernelSubset

        if isinstance(iid_indexer_and_snp_indexer, tuple):
            iid0_indexer, iid1_indexer = iid_indexer_and_snp_indexer
        else:
            iid0_indexer = iid_indexer_and_snp_indexer
            iid1_indexer = iid0_indexer

        return _KernelSubset(self, iid0_indexer, iid1_indexer)

    def _assert_iid0_iid1(self, check_val):
        if check_val:
            assert len(self._val.shape) == 2, "val should have two dimensions"
            assert self._val.shape == (
                len(self._row),
                len(self._col),
            ), "val shape should match that of iid0_count x iid1_count"

        assert (
            self._row.dtype.type is np.str_
            and len(self._row.shape) == 2
            and self._row.shape[1] == 2
        ), "iid0 should be dtype str, have two dimensions, and the second dimension should be size 2"
        assert (
            self._col.dtype.type is np.str_
            and len(self._col.shape) == 2
            and self._col.shape[1] == 2
        ), "iid1 should be dtype str have two dimensions, and the second dimension should be size 2"

    def _read_with_standardizing(
        self,
        to_kerneldata,
        snp_standardizer=None,
        kernel_standardizer=DiagKtoN(),
        return_trained=False,
    ):
        assert (
            to_kerneldata
        ), "When working with non-SnpKernels, to_kerneldata must be 'True'"
        kernel, kernel_trained = self.read().standardize(
            kernel_standardizer, return_trained=True
        )

        if return_trained:
            return kernel, None, kernel_trained
        else:
            return kernel

    @property
    def val_shape(self):
        """
        Tells the shape of value for a given individual and SNP. For KernelReaders always returns None, meaning a single scalar value.
        """
        return None


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test
    print("done")



#!!! Has not been tested

import logging
import os
import numpy as np
from pysnptools.pstreader import PstReader


class _MergeRows(PstReader):  # !!!why does this start with _
    def __init__(self, reader_list, cache_file=None, skip_check=False):
        super(_MergeRows, self).__init__()
        assert len(reader_list) > 0, "Expect at least one reader"
        self.skip_check = skip_check

        self.reader_list = reader_list
        self._repr_string = "_MergeRows({0})".format(
            ",".join([str(s) for s in reader_list])
        )

        if cache_file is not None:
            if not os.path.exists(cache_file):
                self._run_once()
                np.savez(
                    cache_file,
                    _row=self._row,
                    _row_property=self._row_property,
                    _row_count_list=self._row_count_list,
                )
            else:
                with np.load(cache_file, allow_pickle=True) as data:
                    self._row = data["_row"]
                    self._row_property = data["_row_property"]
                    self._row_count_list = data["_row_count_list"]
                    self._has_run_once = True

    def _run_once(self):
        if hasattr(self, "_has_run_once"):
            return
        self._has_run_once = True
        # Check that all iids are distinct and that all sids and pos are the same and in the same order

        row_list = []
        row_property_list = []
        row_set = set()
        col = self.reader_list[0].col
        col_property = self.reader_list[0].col_property
        for reader_index, reader in enumerate(self.reader_list):
            if reader_index % 10 == 0:
                logging.info(
                    "_MergeRows looking at reader #{0}: {1}".format(
                        reader_index, reader
                    )
                )
            if not self.skip_check:
                assert np.array_equal(
                    col, reader.col
                ), "Expect columns to be the same across all files"
                np.testing.assert_equal(
                    col_property, reader.col_property
                )  # "Expect column_property to be the same across all files"
                size_before = len(row_set)
                row_set.update((tuple(item) for item in reader.row))
                assert (
                    len(row_set) == size_before + reader.row_count
                ), "Expect rows to be distinct in all files"
            row_list.append(reader.row)
            row_property_list.append(reader.row_property)
        self._row = np.concatenate(row_list)
        self._row_property = np.concatenate(row_property_list)
        self._row_count_list = [len(row) for row in row_list]

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self.reader_list[0].col

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    @property
    def col_property(self):
        self._run_once()
        return self.reader_list[0].col_property

    def __repr__(self):
        # Don't need _run_once because based only on initial info
        return self._repr_string

    def copyinputs(self, copier):
        self._run_once()
        for reader in self.reader_list:
            copier.input(reader)

    def _create_reader_and_row_index_list(self, row_index):
        result = []
        start = 0
        for reader_index in range(len(self.reader_list)):  # !!!this needs test coverage
            stop = start + self._row_count_list[reader_index]
            is_here = (row_index >= start) * (row_index < stop)
            if any(is_here):
                row_index_rel = row_index[is_here] - start
                result.append((reader_index, is_here, row_index_rel))
            start = stop
        return result

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        # !!!tests to do: no row's
        # !!!tests to do: no col's
        # !!!test to do: from file 1, file2, and then file1 again
        dtype = np.dtype(dtype)
        row_index = (
            row_index_or_none
            if row_index_or_none is not None
            else np.arange(self.row_count)
        )  # !!!might want to special case reading all
        col_index_or_none_count = (
            self.col_count if col_index_or_none is None else len(col_index_or_none)
        )
        reader_and_row_index_list = self._create_reader_and_row_index_list(row_index)

        if len(reader_and_row_index_list) == 0:
            return self.reader_list[0]._read(
                row_index,
                col_index_or_none,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )
        elif len(reader_and_row_index_list) == 1:
            reader_index, row_index_in, row_index_rel = reader_and_row_index_list[0]
            reader = self.reader_list[reader_index]
            return reader._read(
                row_index_rel,
                col_index_or_none,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )
        else:
            logging.info(
                "Starting read from {0} subreaders".format(
                    len(reader_and_row_index_list)
                )
            )
            if order == "A" or order is None:
                order = "F"
            val = np.empty(
                (len(row_index), col_index_or_none_count), dtype=dtype, order=order
            )
            for reader_index, is_here, row_index_rel in reader_and_row_index_list:
                reader = self.reader_list[reader_index]
                if reader_index % 1 == 0:
                    logging.info("Reading from #{0}: {1}".format(reader_index, reader))
                val[is_here, :] = reader._read(
                    row_index_rel,
                    col_index_or_none,
                    order,
                    dtype,
                    force_python_only,
                    view_ok=True,
                    num_threads=num_threads,
                )
            logging.info(
                "Ended read from {0} subreaders".format(len(reader_and_row_index_list))
            )
            return val



'''
.. deprecated::
    Use :class:`.Pheno` instead.    
'''

import logging
import sys
import scipy as sp
import numpy as np

def loadOnePhen(filename,  i_pheno = 0, missing = '-9', vectorize = False):
    '''
    .. deprecated::
       Use :class:`.Pheno` instead.    

    Load one column of a phenotype file. Remove any rows with missing data

    :param filename: name of the file
    :type filename: string
    :param i_pheno: column to return (default '0', the first column)
    :type i_pheno: int
    :param missing: value to threat as missing
    :type missing: string
    :param vectorize: if true, return a 1-D vector rather than a 2-D array
    :type vectorize: bool

    :rtype: An output dictionary

    The output dictionary looks like:

    * 'header' : [1] array phenotype namesv (only if header line is specified in file),
    * 'vals'   : [N*1] array of phenotype-data,
    * 'iid'    : [N*2] array of family IDs and individual IDs
    '''
    missing = missing
    allColumns = loadPhen(filename, missing)
    i_present=allColumns['vals'][:,i_pheno]==allColumns['vals'][:,i_pheno]
    valsvector = allColumns['vals'][i_present,i_pheno]
    vals = np.reshape(valsvector,(-1,1))
    iid = allColumns['iid'][i_present,:]
    #iid = iid.reshape(iid.shape[1], iid.shape[2])
    header = allColumns['header']
    if header is not None:
        header = [header[i_pheno]]

    if vectorize:
        vals = vals[:,0]

    ret = {
            'header':header,
            'vals':vals,
            'iid':iid
            }
    return ret


def loadPhen(filename, missing = '-9',famid='FID', sampid='ID'):
    '''
    .. deprecated::
       Use :class:`.Pheno` instead.    

    Load a phenotype or covariate file. Covariates have the same file format.

    :param filename: name of the file
    :type filename: string
    :param missing: value to threat as missing
    :type missing: string
    :param vectorize: if true, return a 1-D vector rather than a 2-D array
    :type vectorize: bool

    :rtype: An output dictionary

    The output dictionary looks like:

    * 'header' : [1] array phenotype names (only if header line is specified in file),
    * 'vals'   : [N*1] array of phenotype-data,
    * 'iid'    : [N*2] array of family IDs and individual IDs
    '''
    if missing == '-9':
        logging.warning("loadPhen is using default missing value of '-9'.")

    data = np.loadtxt(filename, dtype='str', comments=None)
    data = data.reshape(-1,data.shape[-1]) #Turns 1-d row into 2-d
    if data[0,0] == sampid: #One column of ids - use the single id as both the family id and the iid
        header = data[0,1::].tolist()
        iid = data[1:,[0,0]]
        valsStr = data[1:,1:]
    elif data[0,0] == famid:
        header = data[0,2::].tolist()
        iid = data[1:,0:2]
        valsStr = data[1:,2:]
    else:
        header = [None] * (data.shape[1]-2) # create a header containing a list of None's
        iid = data[:,0:2]
        valsStr = data[:,2:]

    
    if missing is not None:
        valsStr[valsStr==missing] = "NaN"
    vals = np.array(valsStr,dtype = 'float')

    ret = {
            'header':header,
            'vals':vals,
            'iid':iid
            }
    return ret

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import logging
import os
from pathlib import Path
import platform
from shutil import copyfile

import numpy as np
import pytest
from numpy import array, array_equal, isnan
from numpy.testing import assert_allclose, assert_equal

from bgen_reader import open_bgen
from bgen_reader._environment import BGEN_READER_CACHE_HOME
from bgen_reader.test.write_random import _write_random
from bgen_reader.test.test_bgen_reader import noread_permission

from pysnptools.util import example_file_bgen


def test_bgen_samples_inside_bgen():
    data = open_bgen(example_file_bgen("haplotypes.bgen"), verbose=False)
    samples = ["sample_0", "sample_1", "sample_2", "sample_3"]
    assert all(data.samples == samples)


def test_typing():
    data = open_bgen(example_file_bgen("haplotypes.bgen"), verbose=False)
    with pytest.raises(TypeError):
        data.read(dtype=3)


def test_bgen_samples_not_present():
    data = open_bgen(
        example_file_bgen("complex.23bits.no.samples.bgen"),
        allow_complex=True,
        verbose=False,
    )
    samples = ["sample_0", "sample_1", "sample_2", "sample_3"]
    assert all(data.samples == samples)


def test_bgen_samples_specify_samples_file():
    data = open_bgen(
        example_file_bgen("complex.23bits.bgen"),
        samples_filepath=example_file_bgen("complex.sample"),
        allow_complex=True,
        verbose=False,
    )
    samples = ["sample_0", "sample_1", "sample_2", "sample_3"]
    assert all(data.samples == samples)


# skip this test completely
# @pytest.mark.skipif(platform.system() != "Darwin", reason="only reliable on macos")
# def test_bgen_samples_outside_bgen_unreadable(tmp_path):
#     bgen_filepath = example_file_bgen("complex.23bits.bgen")
#     samples_filepath = tmp_path / "complex.sample"
#     copyfile(example_file_bgen("complex.sample"), samples_filepath)
#     with noread_permission(samples_filepath):
#         with pytest.raises(PermissionError):
#             open_bgen(
#                 bgen_filepath,
#                 samples_filepath=samples_filepath,
#                 allow_complex=True,
#                 verbose=False,
#             )


@pytest.mark.skipif(platform.system() != "Darwin", reason="only reliable on macos")
def test_bgen_file_not_readable(tmp_path):
    filepath = tmp_path / "haplotypes.bgen"
    copyfile(example_file_bgen("haplotypes.bgen"), filepath)
    with noread_permission(filepath):
        with pytest.raises(PermissionError):
            open_bgen(filepath, verbose=False)


def test_bgen_file_dont_exist():
    with pytest.raises(FileNotFoundError):
        open_bgen("idontexist.bgen", verbose=False)


def test_metafile_not_provided():
    open_bgen(example_file_bgen("haplotypes.bgen"), verbose=False)


def test_open_bgen_phased_genotype():
    filepath = example_file_bgen("haplotypes.bgen")
    bgen2 = open_bgen(filepath, verbose=False)

    assert_equal(bgen2.chromosomes[0], "1")
    assert_equal(bgen2.ids[0], "SNP1")
    assert_equal(bgen2.nalleles[0], 2)
    assert_equal(bgen2.allele_ids[0], "A,G")
    assert_equal(bgen2.positions[0], 1)
    assert_equal(bgen2.rsids[0], "RS1")

    assert_equal(bgen2.chromosomes[2], "1")
    assert_equal(bgen2.ids[2], "SNP3")
    assert_equal(bgen2.nalleles[2], 2)
    assert_equal(bgen2.allele_ids[2], "A,G")
    assert_equal(bgen2.positions[2], 3)
    assert_equal(bgen2.rsids[2], "RS3")

    assert_equal(bgen2.samples[0], "sample_0")
    assert_equal(bgen2.samples[2], "sample_2")
    assert_equal(bgen2.samples[-1], "sample_3")

    g = bgen2.read((0, 0))
    assert_allclose(g[0, 0, :], [1.0, 0.0, 1.0, 0.0])
    g = bgen2.read((-1, -1))
    assert_allclose(g[0, 0, :], [1.0, 0.0, 0.0, 1.0])


def test_open_bgen_variants_info():
    filepath = example_file_bgen("example.32bits.bgen")
    bgen2 = open_bgen(filepath, verbose=False)

    assert_equal(bgen2.chromosomes[0], "01")
    assert_equal(bgen2.ids[0], "SNPID_2")
    assert_equal(bgen2.nalleles[0], 2)
    assert_equal(bgen2.allele_ids[0], "A,G")
    assert_equal(bgen2.positions[0], 2000)
    assert_equal(bgen2.rsids[0], "RSID_2")

    assert_equal(bgen2.chromosomes[7], "01")
    assert_equal(bgen2.ids[7], "SNPID_9")
    assert_equal(bgen2.nalleles[7], 2)
    assert_equal(bgen2.allele_ids[7], "A,G")
    assert_equal(bgen2.positions[7], 9000)
    assert_equal(bgen2.rsids[7], "RSID_9")

    assert_equal(bgen2.chromosomes[-1], "01")
    assert_equal(bgen2.ids[-1], "SNPID_200")
    assert_equal(bgen2.nalleles[-1], 2)
    assert_equal(bgen2.allele_ids[-1], "A,G")
    assert_equal(bgen2.positions[-1], 100001)
    assert_equal(bgen2.rsids[-1], "RSID_200")

    assert_equal(bgen2.samples[0], "sample_001")
    assert_equal(bgen2.samples[7], "sample_008")
    assert_equal(bgen2.samples[-1], "sample_500")

    g = bgen2.read((0, 0))
    assert all(isnan(g[0, 0, :]))

    g = bgen2.read((1, 0))
    a = [0.027802362811705648, 0.00863673794284387, 0.9635608992454505]
    assert_allclose(g[0, 0, :], a)

    b = [
        0.97970582847010945215516,
        0.01947019668749305418287,
        0.00082397484239749366197,
    ]
    g = bgen2.read((2, 1))
    assert_allclose(g[0, 0, :], b)


def test_to_improve_coverage():
    filepath = Path(example_file_bgen("example.32bits.bgen"))
    bgen2 = open_bgen(filepath, verbose=False)  # Creates metadata2.npz file
    assert_equal(bgen2.ncombinations[-1], 3)
    assert_equal(bgen2.phased[-1], False)
    with open_bgen(filepath) as bgen2:  # Reuses metadata2.npz file
        assert_equal(str(bgen2), "open_bgen('{0}')".format(filepath.name))
        assert_equal(bgen2.nsamples, 500)
        assert_equal(bgen2.nvariants, 199)
        assert_equal(bgen2.shape, (500, 199, 3))
        assert_equal(bgen2.ids[-1], "SNPID_200")
        assert_equal(bgen2.rsids[-1], "RSID_200")
        assert_equal(bgen2.chromosomes[-1], "01")
        assert_equal(bgen2.positions[-1], 100001)
        assert_equal(bgen2.nalleles[-1], 2)
        assert_equal(bgen2.allele_ids[-1], "A,G")
        assert_equal(bgen2.ncombinations[-1], 3)
        assert_equal(bgen2.phased[-1], False)
        assert_equal(bgen2.samples[-1], "sample_500")

        b = [
            0.97970582847010945215516,
            0.01947019668749305418287,
            0.00082397484239749366197,
        ]
        g = bgen2.read((2, 1))
        assert_allclose(g[0, 0, :], b)

        g = bgen2.read()
        assert_allclose(g[2, 1, :], b)

    # confirm that out-of-date metadata2 file will be updated
    metadata2 = open_bgen._metadata_path_from_filename(filepath, samples_filepath=None)
    assert os.path.getmtime(metadata2) >= os.path.getmtime(filepath)
    filepath.touch()
    assert os.path.getmtime(metadata2) <= os.path.getmtime(filepath)
    bgen2 = open_bgen(filepath, verbose=False)  # Creates metadata2.npz file
    assert os.path.getmtime(metadata2) >= os.path.getmtime(filepath)


# @pytest.mark.skipif(
#     "QCTOOLPATH" not in os.environ, reason="This test requires external QCTOOL"
# )
# @pytest.mark.slow  # It takes hours to generate data locally.  After that, it takes a few minutes
# # to run.
# def test_bigfile(verbose=False):
#     random_file_tests(nsamples=2500, nvariants=500 * 1000, bits=16)


# @pytest.mark.skipif(
#     "QCTOOLPATH" not in os.environ, reason="This test requires external QCTOOL"
# )
# @pytest.mark.slow  # Skipping this one by default because it requires the QCTOOL
# def test_small_random_file(verbose=False):
#     random_file_tests(
#         nsamples=25, nvariants=1000, bits=8, verbose=verbose, overwrite=True
#     )


def random_file_tests(nsamples, nvariants, bits, verbose=False, overwrite=False):
    test_data_folder = BGEN_READER_CACHE_HOME / "test_data"
    filepath = test_data_folder / "{0}x{1}.{2}bits.bgen".format(
        nsamples, nvariants, bits
    )
    if overwrite or not filepath.exists():
        _write_random(
            filepath,
            nsamples,
            nvariants,
            bits=bits,
            verbose=verbose,
            cleanup_temp_files=True,
        )
    metadata2_path = open_bgen._metadata_path_from_filename(
        filepath, samples_filepath=None
    )
    if metadata2_path.exists():
        metadata2_path.unlink()

    with open_bgen(filepath, verbose=verbose) as bgen2:
        assert bgen2.nsamples == nsamples
        assert bgen2.nvariants == nvariants
        val = bgen2.read(-1)
        assert val.shape == (nsamples, 1, 3)
        mean = np.nanmean(val)
        assert mean != mean or 0 <= mean <= 1


def test_open_bgen_without_metadata():
    filepath = example_file_bgen("example.32bits.bgen")
    bgen2 = open_bgen(filepath, verbose=False)
    assert_equal(bgen2.allele_ids[7], "A,G")
    assert_equal(bgen2.samples[-1], "sample_500")


def test_open_bgen_file_notfound():
    with pytest.raises(FileNotFoundError):
        open_bgen("/1/2/3/example.32bits.bgen", verbose=False)


def test_open_bgen_complex():
    filepath = example_file_bgen("complex.23bits.bgen")
    bgen2 = open_bgen(filepath, allow_complex=True, verbose=False)

    assert_equal(bgen2.chromosomes[0], "01")
    assert_equal(bgen2.ids[0], "")
    assert_equal(bgen2.nalleles[0], 2)
    assert_equal(bgen2.allele_ids[0], "A,G")
    assert_equal(bgen2.positions[0], 1)
    assert_equal(bgen2.rsids[0], "V1")

    assert_equal(bgen2.chromosomes[7], "01")
    assert_equal(bgen2.ids[7], "")
    assert_equal(bgen2.nalleles[7], 7)
    assert_equal(bgen2.allele_ids[7], "A,G,GT,GTT,GTTT,GTTTT,GTTTTT")
    assert_equal(bgen2.positions[7], 8)
    assert_equal(bgen2.rsids[7], "M8")

    assert_equal(bgen2.chromosomes[-1], "01")
    assert_equal(bgen2.ids[-1], "")
    assert_equal(bgen2.nalleles[-1], 2)
    assert_equal(bgen2.allele_ids[-1], "A,G")
    assert_equal(bgen2.positions[-1], 10)
    assert_equal(bgen2.rsids[-1], "M10")

    assert_equal(bgen2.samples[0], "sample_0")
    assert_equal(bgen2.samples[3], "sample_3")

    g = bgen2.read((0, 0))
    assert_allclose(g[0, 0, :2], [1, 0])
    assert isnan(g[0, 0, 2])

    g = bgen2.read((1, 0))
    assert_allclose(g[0, 0, :3], [1, 0, 0])

    g = bgen2.read((-1, -1))
    assert_allclose(g[0, 0, :5], [0, 0, 0, 1, 0])

    ploidy = bgen2.read(0, return_probabilities=False, return_ploidies=True)
    assert_allclose(ploidy[:, 0], [1, 2, 2, 2])
    ploidy = bgen2.read(-1, return_probabilities=False, return_ploidies=True)
    assert_allclose(ploidy[:, 0], [4, 4, 4, 4])

    assert_equal(bgen2.phased.dtype.name, "bool")
    ideal = array([False, True, True, False, True, True, True, True, False, False])
    assert array_equal(bgen2.phased, ideal)


def test_open_bgen_complex_sample_file():
    bgen2 = open_bgen(
        example_file_bgen("complex.23bits.bgen"),
        samples_filepath=example_file_bgen("complex.sample"),
        allow_complex=True,
        verbose=False,
    )

    assert_equal(bgen2.chromosomes[0], "01")
    assert_equal(bgen2.ids[0], "")
    assert_equal(bgen2.nalleles[0], 2)
    assert_equal(bgen2.allele_ids[0], "A,G")
    assert_equal(bgen2.positions[0], 1)
    assert_equal(bgen2.rsids[0], "V1")

    assert_equal(bgen2.chromosomes[7], "01")
    assert_equal(bgen2.ids[7], "")
    assert_equal(bgen2.nalleles[7], 7)
    assert_equal(bgen2.allele_ids[7], "A,G,GT,GTT,GTTT,GTTTT,GTTTTT")
    assert_equal(bgen2.positions[7], 8)
    assert_equal(bgen2.rsids[7], "M8")

    assert_equal(bgen2.chromosomes[-1], "01")
    assert_equal(bgen2.ids[-1], "")
    assert_equal(bgen2.nalleles[-1], 2)
    assert_equal(bgen2.allele_ids[-1], "A,G")
    assert_equal(bgen2.positions[-1], 10)
    assert_equal(bgen2.rsids[-1], "M10")

    assert_equal(bgen2.samples[0], "sample_0")
    assert_equal(bgen2.samples[3], "sample_3")

    missing, ploidy = bgen2.read(
        2, return_probabilities=False, return_missings=True, return_ploidies=True
    )
    assert_allclose(ploidy[:, 0], [1, 2, 2, 2])
    assert_allclose(missing[:, 0], [0, 0, 0, 0])
    assert_allclose(bgen2.phased, [0, 1, 1, 0, 1, 1, 1, 1, 0, 0])


def test_close_del_with():
    filepath = example_file_bgen("example.32bits.bgen")
    with open_bgen(filepath, verbose=False) as bgen2:
        pass
    with pytest.raises(ValueError):
        bgen2.read()

    bgen2 = open_bgen(filepath, verbose=False)
    bgen2.close()
    with pytest.raises(AttributeError):
        bgen2.samples
    with pytest.raises(ValueError):
        bgen2.read()


def test_read_max_combinations():
    filepath = example_file_bgen("example.32bits.bgen")
    with open_bgen(filepath, verbose=False) as bgen2:
        assert (
            np.mean(np.isnan(bgen2.read())) < 2.1e-05
        )  # main data as only a few missing
        val = bgen2.read(max_combinations=5)
        assert (
            np.mean(np.isnan(val[:, :, :3])) < 2.1e-05
        )  # main data as only a few missing
        assert np.all(np.isnan(val[:, :, 3:]))  # all the extra are NaN
        with pytest.raises(ValueError):
            bgen2.read(max_combinations=2)
        with pytest.raises(ValueError):
            bgen2.read(max_combinations=1)
        with pytest.raises(ValueError):
            bgen2.read(max_combinations=0)


def test_read_dtype_and_order():
    filepath = example_file_bgen("example.32bits.bgen")
    bgen2 = open_bgen(filepath, verbose=False)
    full = bgen2.read()
    assert full.dtype == np.float64
    assert full.flags["F_CONTIGUOUS"] and not full.flags["C_CONTIGUOUS"]

    val = bgen2.read(None, dtype="float32", order="C")
    assert val.dtype == np.float32
    assert val.flags["C_CONTIGUOUS"] and not val.flags["F_CONTIGUOUS"]
    assert np.allclose(full, val, atol=1e-7, equal_nan=True)


def test_read_indexing():
    filepath = example_file_bgen("example.32bits.bgen")
    bgen2 = open_bgen(filepath, verbose=False)
    full = bgen2.read()

    val = bgen2.read(22)
    assert np.allclose(full[:, [22]], val, equal_nan=True)

    val = bgen2.read([22])
    assert np.allclose(full[:, [22]], val, equal_nan=True)

    val = bgen2.read([22, 30])
    assert np.allclose(full[:, [22, 30]], val, equal_nan=True)

    val = bgen2.read(slice(10, 30, 2))
    assert np.allclose(full[:, 10:30:2], val, equal_nan=True)

    bool_list = [i % 2 == 0 for i in range(bgen2.nvariants)]
    val = bgen2.read(bool_list)
    assert np.allclose(full[:, bool_list], val, equal_nan=True)

    val = bgen2.read((None, None))
    assert np.allclose(full, val, equal_nan=True)

    val = bgen2.read((22, None))
    assert np.allclose(full[[22], :], val, equal_nan=True)

    val = bgen2.read((22, [11, 9]))
    assert np.allclose(full[[22], :][:, [11, 9]], val, equal_nan=True)

    val = bgen2.read(([22, 30], [11, 9]))
    assert np.allclose(full[[22, 30], :][:, [11, 9]], val, equal_nan=True)

    val = bgen2.read((slice(10, 30, 2), [11, 9]))
    assert np.allclose(full[10:30:2, :][:, [11, 9]], val, equal_nan=True)

    bool_list = [i % 2 == 0 for i in range(bgen2.nsamples)]
    val = bgen2.read((bool_list, [11, 9]))
    assert np.allclose(full[bool_list, :][:, [11, 9]], val, equal_nan=True)

    val = bgen2.read(([-1], [-1]))
    assert np.allclose(full[-1, -1], val, equal_nan=True)

    val = bgen2.read(np.s_[10:30:2, :5])
    assert np.allclose(full[10:30:2, :5, :], val, equal_nan=True)

    # Read no variants
    val, missing, ploidy = bgen2.read([], return_missings=True, return_ploidies=True)
    assert val.shape == (bgen2.nsamples, 0, bgen2.max_combinations)
    assert missing.shape == (bgen2.nsamples, 0)
    assert ploidy.shape == (bgen2.nsamples, 0)

    # Read no samples and no variants
    val, missing, ploidy = bgen2.read(
        ([], []), return_missings=True, return_ploidies=True
    )
    assert val.shape == (0, 0, bgen2.max_combinations)
    assert missing.shape == (0, 0)
    assert ploidy.shape == (0, 0)


def test_read_multiple_returns():
    filepath = example_file_bgen("example.32bits.bgen")
    bgen2 = open_bgen(filepath, verbose=False)
    full, full_missing, full_ploidy = bgen2.read(
        return_missings=True, return_ploidies=True
    )

    val, missing = bgen2.read(return_missings=True)
    assert np.allclose(full, val, equal_nan=True)
    assert np.allclose(full_missing, missing, equal_nan=False)

    ploidy = bgen2.read(return_probabilities=False, return_ploidies=True)
    assert np.allclose(full_ploidy, ploidy, equal_nan=False)

    val, missing = bgen2.read((slice(10, 30, 2), [11, 9]), return_missings=True)
    assert np.allclose(full[10:30:2, :][:, [11, 9]], val, equal_nan=True)
    assert np.allclose(full_missing[10:30:2, :][:, [11, 9]], missing, equal_nan=False)

    ploidy = bgen2.read(
        (slice(10, 30, 2), [11, 9]), return_probabilities=False, return_ploidies=True
    )
    assert np.allclose(full_ploidy[10:30:2, :][:, [11, 9]], ploidy, equal_nan=False)


def test_respect_read_inputs():
    from pysnptools.distreader import (
        DistGen,
        DistHdf5,
        DistMemMap,
        DistNpz,
        _DistMergeSIDs,
        Bgen,
    )
    from pysnptools.snpreader import Bed

    previous_wd = os.getcwd()
    os.chdir(os.path.dirname(os.path.realpath(__file__)))

    distreader_list = [
        Bed("../examples/toydata.5chrom.bed", count_A1=True).as_dist(block_size=2000),
        Bed("../examples/toydata.5chrom.bed", count_A1=True).as_dist(),
        DistGen(seed=0, iid_count=500, sid_count=50),
        DistGen(seed=0, iid_count=500, sid_count=50)[::2, ::2],
        DistHdf5("../examples/toydata.snpmajor.dist.hdf5"),
        DistMemMap("../examples/tiny.dist.memmap"),
        DistNpz("../examples/toydata10.dist.npz"),
    ]
    distreader_list += [
        _DistMergeSIDs(
            [
                Bgen("../examples/example.bgen")[:, :5].read(),
                Bgen("../examples/example.bgen")[:, 5:].read(),
            ]
        ),
        Bgen("../examples/example.bgen").read(),
        Bgen("../examples/bits1.bgen"),
    ]

    for distreader in distreader_list:
        logging.info(str(distreader))
        for order in ["F", "C", "A"]:
            for dtype in [np.float32, np.float64]:
                for force_python_only in [True, False]:
                    for view_ok in [True, False]:
                        val = distreader.read(
                            order=order,
                            dtype=dtype,
                            force_python_only=force_python_only,
                            view_ok=view_ok,
                        ).val
                        has_right_order = (
                            order == "A"
                            or (order == "C" and val.flags["C_CONTIGUOUS"])
                            or (order == "F" and val.flags["F_CONTIGUOUS"])
                        )
                        if hasattr(distreader, "val") and not view_ok:
                            assert distreader.val is not val
                        if (
                            hasattr(distreader, "val")
                            and view_ok
                            and distreader.val is not val
                            and (
                                order == "A"
                                or (
                                    order == "F"
                                    and distreader.val.flags["F_CONTIGUOUS"]
                                )
                                or (
                                    order == "C"
                                    and distreader.val.flags["C_CONTIGUOUS"]
                                )
                            )
                            and (dtype is None or distreader.val.dtype == dtype)
                        ):
                            logging.info(
                                "{0} could have read a view, but didn't".format(
                                    distreader
                                )
                            )
                        assert val.dtype == dtype and has_right_order

    os.chdir(previous_wd)


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import absolute_import
import os
import shutil
import numpy as np
import logging
import unittest
from pysnptools.snpreader import _MergeSIDs
from pysnptools.snpreader import SnpReader, Bed
from pysnptools.pstreader import PstReader
from pysnptools.util import _multiopen
from pysnptools.snpreader import _snps_fixup
from pysnptools.util import log_in_place
from pysnptools.util.mapreduce1 import map_reduce
from pysnptools.util.filecache import FileCache


class DistributedBed(SnpReader):
    """
    A class that implements the :class:`SnpReader` interface. It stores :class:`.Bed`-like data in pieces on storage. When you request data, it retrieves only the needed pieces.

    **Constructor:**
        :Parameters: **storage** (string or :class:`.FileCache`) -- Tells where the DistirubtedBed data is stored.
                      A string can be given and will be interpreted as the path to a directory.
                      A :class:`.FileCache` instance can be given, which provides a method to specify cluster-distributed storage.
        :type storage: string or :class:`.FileCache`

        :Example:

        >>> import os
        >>> from pysnptools.snpreader import DistributedBed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> folder = os.path.dirname(example_file('pysnptools/examples/toydataSkip10.distributedbed/*'))
        >>> data_on_disk = DistributedBed(folder)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 1000)

    """

    def __init__(self, storage):
        super(DistributedBed, self).__init__()

        self._ran_once = False
        self._storage = FileCache._fixup(storage)
        self._merge = None

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, self._storage)

    @property
    def row(self):
        """*same as* :attr:`iid`"""
        self._run_once()
        return self._merge.row

    @property
    def col(self):
        """*same as* :attr:`sid`"""
        self._run_once()
        return self._merge.col

    @property
    def col_property(self):
        """*same as* :attr:`pos`"""
        self._run_once()
        return self._merge.col_property

    def _run_once(self):
        if self._ran_once:
            return
        self._ran_once = True

        _metadatanpz = "metadata.npz"
        with self._storage.open_read(_metadatanpz) as handle_metadatanpz_file_name:
            # self._file_dict["metadatanpz"] = handle_metadatanpz
            _reader_name_listnpz = "reader_name_list.npz"
            with self._storage.open_read(
                _reader_name_listnpz
            ) as handle_reader_name_listnpz_file_name:
                reader_name_list = np.array(
                    np.load(handle_reader_name_listnpz_file_name)["reader_name_list"],
                    dtype="str",
                )
                # self._file_dict["reader_name_listnpz"] = handle_reader_name_listnpz

                reader_list = [
                    _Distributed1Bed(reader_name, self._storage)
                    for reader_name in reader_name_list
                ]

                self._merge = _MergeSIDs(
                    reader_list,
                    cache_file=handle_metadatanpz_file_name,
                    skip_check=True,
                )

                for reader in reader_list:
                    reader._row = self._merge.row

    def __getstate__(self):
        return self._storage

    def __setstate__(self, state):
        storage = state
        self.__init__(storage)

    def copyinputs(self, copier):
        pass

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()
        dtype = np.dtype(dtype)

        return self._merge._read(
            iid_index_or_none,
            sid_index_or_none,
            order,
            dtype,
            force_python_only,
            view_ok,
            num_threads,
        )

    # !!! in future could make a default for piece_per_chrom_count that made each piece some GB size
    @staticmethod
    def write(
        storage, snpreader, piece_per_chrom_count=1, updater=None, runner=None
    ):  # !!! might want to set pieces_per_chrom such that it is a certain size
        """
        Uploads from any :class:`.Bed`-like data to cluster storage for efficient retrieval later.
        If some of the contents already exists in storage, it skips uploading that part of the contents. (To avoid this behavior,
        clear the storage.)

        :param storage: Tells where to store SNP data.
                      A string can be given and will be interpreted as the path of a local directory to use for storage. (The local
                      directory will **not** be automatically erased and so must be user managed.)
                      A :class:`.FileCache` instance can be given, which provides a
                      method to specify cluster-distributed storage. (:class:`.FileCache`'s will **not** be automatically erased and must be user managed.)
                      If `None`, the storage will be in an automatically-erasing temporary directory. (If the TEMP environment variable is set, Python places the temp directory under it.)

        :type storage: string or :class:`.FileCache` or None.

        :param snpreader: A :class:`.Bed` or other :class:`.SnpReader` with values of 0,1,2, or missing.
            (Note that this differs from most other `write` methods that take a :class:`.SnpData`)
        :type snpreader: :class:`.SnpReader`

        :param piece_per_chrom_count: The number of pieces in which to store the data from each chromosome. Data is split across
            SNPs. For exmple, if `piece_per_chrom_count` is set to 100 and 22 chromosomes are uploaded, then data will be stored in 2200 pieces. Later, when data is requested
            only the pieces necessary for the request will be downloaded to local storage.
        :type piece_per_chrom_count: A number

        :param updater: A single argument function to write logging message to, for example, the function created by :func:`.log_in_place`.
        :type updater: A function or lambda

        :param runner: a :class:`.Runner`, optional: Tells how to run.
            (Note that :class:`.Local` and :class:`.LocalMultProc` are good options.)
            If not given, the function is run locally.
        :type runner: :class:`.Runner`

        :rtype: DistributedBed

        >>> from pysnptools.snpreader import DistributedBed, Bed
        >>> import shutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> directory = 'tempdir/toydataSkip10.distributedbed'
        >>> if os.path.exists(directory):
        ...     shutil.rmtree(directory)
        >>> bedfile = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpreader = Bed(bedfile,count_A1=False)[:,::10]  # Read every 10 snps from Bed format
        >>> DistributedBed.write(directory,snpreader,piece_per_chrom_count=5)  # Write data in DistributedBed format
        DistributedBed(LocalCache('tempdir/toydataSkip10.distributedbed'))


        """
        from pysnptools.util import _file_transfer_reporter
        from pysnptools.util.filecache import FileCache

        count_A1 = True  # Make all these's the same for reading and writing so that nothing will change.
        snpreader = _snps_fixup(snpreader, count_A1=count_A1)

        storage = FileCache._fixup(storage)

        chrom_set = sorted(set(snpreader.pos[:, 0]))
        for chrom in chrom_set:
            assert chrom == chrom and chrom == int(
                chrom
            ), "DistributedBed.write expects all chromosomes to be integers (not '{0}')".format(
                chrom
            )
        with _file_transfer_reporter(
            "DistributedBed.write", size=0, updater=updater
        ) as updater2:

            def mapper_closure(chrom):
                chrom_reader = snpreader[:, snpreader.pos[:, 0] == chrom]

                def nested_closure(piece_per_chrom_index):
                    start = (
                        chrom_reader.sid_count
                        * piece_per_chrom_index
                        // piece_per_chrom_count
                    )
                    stop = (
                        chrom_reader.sid_count
                        * (piece_per_chrom_index + 1)
                        // piece_per_chrom_count
                    )
                    piece_reader = chrom_reader[:, start:stop]
                    _piece_name_list = [
                        "chrom{0}.piece{1}of{2}.{3}".format(
                            int(chrom),
                            piece_per_chrom_index,
                            piece_per_chrom_count,
                            suffix,
                        )
                        for suffix in ["bim", "fam", "bed"]
                    ]
                    exist_list = [
                        storage.file_exists(_piece_name)
                        for _piece_name in _piece_name_list
                    ]
                    if (
                        sum(exist_list) < 3
                    ):  # If all three of the BIM/FAM/BED files are already there, then skip the upload, otherwise do the upload
                        for i in range(
                            3
                        ):  # If one or two of BIM/FAM/BED are there, remove them
                            if exist_list[i]:
                                storage.remove(_piece_name_list[i])
                        _Distributed1Bed.write(
                            _piece_name_list[-1],
                            storage,
                            piece_reader.read(),
                            count_A1=count_A1,
                            updater=updater2,
                        )
                    return _piece_name_list[-1]

                return map_reduce(
                    range(piece_per_chrom_count),
                    mapper=nested_closure,
                )

            list_list_pair = map_reduce(
                chrom_set,
                nested=mapper_closure,
                runner=runner,
            )

        reader_name_list = []
        reader_list = []
        for chrom_result in list_list_pair:
            for _piece_name in chrom_result:
                reader_name_list.append(_piece_name)
                reader_list.append(_Distributed1Bed(_piece_name, storage))

        _metadatanpz = "metadata.npz"
        with storage.open_write(_metadatanpz) as local_metadatanpz:
            _reader_name_listnpz = "reader_name_list.npz"
            with storage.open_write(_reader_name_listnpz) as local_reader_name_listnpz:
                reader_name_list_ascii = np.array(reader_name_list, dtype="S")
                np.savez(
                    local_reader_name_listnpz, reader_name_list=reader_name_list_ascii
                )
                if os.path.exists(local_metadatanpz):
                    os.remove(local_metadatanpz)
                _MergeSIDs(reader_list, cache_file=local_metadatanpz, skip_check=True)

        return DistributedBed(storage)


class _Distributed1Bed(SnpReader):
    """
    An atomic set of bed/bim/fam files stored somewhere. Can answer metadata questions without downloading the *.bed file.
    But does download the whole *.bed file when any SNP value is requested.
    """

    def __init__(self, path, storage):
        super(_Distributed1Bed, self).__init__()

        self._ran_once = False
        self._file_dict = {}

        self._storage = storage
        self.path = path
        self.local = None

    def __repr__(self):
        return "{0}('{1}','{2}')".format(
            self.__class__.__name__, self.path, self._storage
        )

    @property
    def row(self):
        """*same as* :attr:`iid`"""
        if not hasattr(self, "_row"):
            _fam = SnpReader._name_of_other_file(
                self.path, remove_suffix="bed", add_suffix="fam"
            )
            local_fam = self._storage.open_read(_fam)
            self._row = SnpReader._read_fam(local_fam.__enter__(), remove_suffix="fam")
            self._file_dict["fam"] = local_fam
        return self._row

    @property
    def col(self):
        """*same as* :attr:`sid`"""
        if not hasattr(self, "_col"):
            _bim = SnpReader._name_of_other_file(
                self.path, remove_suffix="bed", add_suffix="bim"
            )
            local_bim = self._storage.open_read(_bim)
            self._col, self._col_property = SnpReader._read_map_or_bim(
                local_bim.__enter__(), remove_suffix="bim", add_suffix="bim"
            )
            self._file_dict["bim"] = local_bim
        return self._col

    @property
    def col_property(self):
        """*same as* :attr:`pos`"""
        if not hasattr(self, "_col"):
            self.col  # get col info
        return self._col_property

    def _run_once(self):
        if self._ran_once:
            return
        self._ran_once = True
        self.row  # get row info
        self.col  # get col info

        _bed = SnpReader._name_of_other_file(
            self.path, remove_suffix="bed", add_suffix="bed"
        )
        local_bed = self._storage.open_read(_bed)
        self.local = Bed(
            local_bed.__enter__(),
            count_A1=True,
            iid=self.row,
            sid=self.col,
            pos=self.col_property,
            skip_format_check=True,
        )
        self._file_dict["bed"] = local_bed

    def __del__(self):
        for handle in self._file_dict.values():
            handle.__exit__(None, None, None)
        self._file_dict = {}

    def copyinputs(self, copier):
        pass

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()
        dtype = np.dtype(dtype)

        return self.local._read(
            iid_index_or_none,
            sid_index_or_none,
            order,
            dtype,
            force_python_only,
            view_ok,
            num_threads,
        )

    @staticmethod
    def write(path, storage, snpdata, count_A1=True, updater=None):
        file_list = [
            SnpReader._name_of_other_file(
                path, remove_suffix="bed", add_suffix=new_suffix
            )
            for new_suffix in ["bim", "fam", "bed"]
        ]  #'bed' should be last
        with _multiopen(
            lambda file_name: storage.open_write(file_name, updater=updater), file_list
        ) as local_file_name_list:
            Bed.write(local_file_name_list[-1], snpdata, count_A1=count_A1)

        return _Distributed1Bed(path, storage)


class TestDistributedBed(unittest.TestCase):
    def test1(self):
        logging.info("in TestDistributedBed test1")
        from pysnptools.snpreader import SnpGen, DistributedBed

        snpgen = SnpGen(seed=0, iid_count=100, sid_count=100)

        temp_dir = "tempdir/distributed_bed_test1"
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)
        distributed_bed = DistributedBed.write(
            temp_dir, snpgen, piece_per_chrom_count=2
        )
        snpdata = distributed_bed.read()

        ref1 = DistributedBed(
            os.path.dirname(os.path.realpath(__file__))
            + "/../../tests/datasets/distributed_bed_test1"
        ).read()
        assert snpdata.allclose(ref1, equal_nan=True)

        ref2 = Bed(
            os.path.dirname(os.path.realpath(__file__))
            + "/../../tests/datasets/distributed_bed_test1_X.bed",
            count_A1=False,
        ).read()
        assert snpdata.allclose(ref2, equal_nan=True)


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistributedBed))
    return test_suite


if __name__ == "__main__":
    import doctest

    logging.basicConfig(level=logging.INFO)

    if False:
        from pysnptools.snpreader import DistributedBed, Bed
        import shutil

        directory = "tempdir/toydataSkip10.distributedbed"
        if os.path.exists(directory):
            shutil.rmtree(directory)
        snpreader = Bed("../examples/toydata.5chrom.bed", count_A1=False)[
            :, ::10
        ]  # Read every 10 snps from Bed format
        DistributedBed.write(
            directory, snpreader, piece_per_chrom_count=5
        )  # Write data in DistributedBed format

    result = doctest.testmod(optionflags=doctest.ELLIPSIS)
    assert result.failed == 0, "failed doc test: " + __file__

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()



import numpy as np
import logging
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData
import warnings
from pysnptools.pstreader import PstData
from pysnptools.pstreader import _OneShot


def zero_family(s):
    """Given an input id from the file, returns 0 as the family id and that input id as the individual id."""
    return "0", s


def zero_pos(s):
    """Given an input id from the file, return the input id as the sid and returns (0,0,0) as the chromosome, genetic distance, basepair distance."""
    return 0, s, 0, 0


def just_case(iid):
    """Use just the individual id (ignoring the family id) in the file."""
    return iid[1]


def just_sid(sid, pos):
    """Use just the sid (ignoring the chromosome, genetic distance, and basepair distance) in the file."""
    return sid


class Dense(_OneShot, SnpReader):
    """
    A :class:`.SnpReader` for reading \*.dense.txt (0,1,2 text files) from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    This format looks like::

        var	4006	9570    22899	37676	41236	41978	55159	66828...
        1-10004-rs12354060	22222222...
        1-707348-rs12184279	222222?2...
        1-724325-rs12564807	00000000...
        ...

    where rows are sids and columns are iids.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The Dense file to read.
                     * **extract_iid_function** (*string*) -- A function that breaks the row id from the file into a family id and an individual id. Defaults to
                       setting the family id to 0 and using the whole row id as the iid.
                     * **extract_sid_pos_function** (*string*) -- A function that breaks the column id from the file into (chromosome, sid, genetic distance, basepair distance).
                       Defaults to setting the :attr:`SnpReader.pos` information to 0 and using the whole column id as the sid.

        :Example:

        >>> from pysnptools.snpreader import Dense
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> dense_file = example_file("pysnptools/examples/toydata100.dense.txt")
        >>> data_on_disk = Dense(dense_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 100)

    **Methods beyond** :class:`.SnpReader`

    """

    def __init__(
        self,
        filename,
        extract_iid_function=zero_family,
        extract_sid_pos_function=zero_pos,
    ):
        """
        filename    : string of the name of the Dat file.
        """
        super(Dense, self).__init__()
        self.filename = str(filename)
        self.extract_iid_function = extract_iid_function
        self.extract_sid_pos_function = extract_sid_pos_function

    def _read_pstdata(self):
        bim_list = []
        val_list_list = []
        with open(self.filename, "r") as fp:
            header = fp.readline()
            iid_string_list = header.strip().split()[1:]
            iid = np.array(
                [
                    self.extract_iid_function(iid_string)
                    for iid_string in iid_string_list
                ],
                dtype="str",
            )
            val_list = []
            zerofloat = float("0"[0])
            missing_char = "?"[0]
            for line_index, line in enumerate(fp):
                if line_index % 1000 == 0:
                    logging.info(
                        "reading sid and iid info from line {0} of file '{1}'".format(
                            line_index, self.filename
                        )
                    )
                sid_string_rest = line.strip().split()
                sid_string = sid_string_rest[0]
                rest = [] if len(sid_string_rest) == 1 else sid_string_rest[1]
                assert len(rest) == len(iid)
                bim_list.append(self.extract_sid_pos_function(sid_string))
                val_list = np.array(
                    [
                        float(val) - zerofloat if val != missing_char else np.NaN
                        for val in rest
                    ]
                )
                val_list_list.append(val_list)

        col = np.array([bim[1] for bim in bim_list], dtype="str")
        col_property = np.array(
            [[bim[0], bim[2], bim[3]] for bim in bim_list], dtype=np.float64
        )

        val = np.zeros((len(iid), len(col)))
        for col_index in range(len(col)):
            val[:, col_index] = val_list_list[col_index]

        return PstData(iid, col, val, col_property=col_property, name=self.filename)

    @staticmethod
    def write(
        filename, snpdata, join_iid_function=just_case, join_sid_pos_function=just_sid
    ):
        """Writes a :class:`SnpData` to Dense format. The values must be 0,1,2 (or missing). Returns a :class:`.Dense` with the default extractors.

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :param join_iid_function: function to turn a family id and individual id into a file id for columns.
                                  Defaults ignoring the family id and using the individual id as the column id.
        :type join_iid_function: a function
        :param join_sid_pos_function: function to turn a sid and pos data into a file id for rows.
                                      Defaults ignoring the :attr:`SnpReader.pos` information and using the sid id as the row id.
        :type join_sid_pos_function: a function
        :rtype: :class:`.Dense`

        >>> from pysnptools.snpreader import Dense, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:,:10].read()  # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.dense.txt")
        >>> Dense.write("tempdir/toydata10.dense.txt",snpdata)        # Write data in Dense format
        Dense('tempdir/toydata10.dense.txt')
        """

        if isinstance(filename, SnpData) and isinstance(
            snpdata, "str"
        ):  # For backwards compatibility, reverse inputs if necessary
            warnings.warn(
                "write statement should have filename before data to write",
                DeprecationWarning,
            )
            filename, snpdata = snpdata, filename

        snpsarray = snpdata.val
        with open(filename, "w") as filepointer:
            filepointer.write("var\t")
            filepointer.write(
                "\t".join((join_iid_function(iid_pair) for iid_pair in snpdata.iid))
                + "\n"
            )

            for sid_index, sid in enumerate(snpdata.sid):
                pos = snpdata.pos[sid_index]
                if sid_index % 1000 == 0:
                    logging.info(
                        "Writing snp # {0} to file '{1}'".format(sid_index, filename)
                    )
                filepointer.write("%s\t" % join_sid_pos_function(sid, pos))
                row = snpsarray[:, sid_index]
                filepointer.write(
                    "".join((str(int(i)) if i == i else "?" for i in row)) + "\n"
                )
        logging.info("Done writing " + filename)
        return Dense(filename)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



# !! Has not been tested

from pysnptools.snpreader import SnpReader
from pysnptools.pstreader import _MergeRows


class _MergeIIDs(_MergeRows, SnpReader):
    def __init__(self, *args, **kwargs):
        super(_MergeIIDs, self).__init__(*args, **kwargs)



from pathlib import Path
import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData
import pysnptools.util.pheno as pstpheno
import pysnptools.util as pstutil
from pysnptools.pstreader import _OneShot


class Pheno(_OneShot, SnpReader):
    """
    A :class:`.SnpReader` for reading "alternative" phenotype files from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    This text format is described `here <http://zzz.bwh.harvard.edu/plink/data.shtml#pheno>`__ and looks like::

         FID    IID      qt1   bmi    site
         F1     1110     2.3   22.22  2
         F2     2202     34.12 18.23  1
         ...

    where the heading row is optional.


    **Constructor:**
        :Parameters: * **input** (*string*) -- The phenotype file to read.
                     * **iid_if_none** (*None* or array of strings) -- An :attr:`SnpReader.iid` to use if the file is empty.
                     * **missing** (*None* or string) -- The value in the file that represents missing data.

        For backwards compatibility with older code, **input** can be a dictionary (instead of a filename) with these fields:

        * 'header' : [1] array phenotype name,
        * 'vals'   : [N*1] array of phenotype data,
        * 'iid'    : [N*2] array of family IDs and individual IDs

        :Example:

        >>> from pysnptools.snpreader import Pheno, Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> pheno_file = example_file('pysnptools/examples/toydata.phe')
        >>> data_on_disk = Pheno(pheno_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 1)

    **Methods beyond** :class:`.SnpReader`

    """

    def __init__(self, input, iid_if_none=None, missing=None):
        """
        input    : string of the name of the file or an in-memory dictionary
        """
        super(Pheno, self).__init__()

        if isinstance(input, Path):
            input = str(input)

        self.filename = input
        self._iid_if_none = iid_if_none
        self.missing = missing

    def _read_pstdata(self):
        # LATER switch it, so the main code is here rather than in loadPhen
        if isinstance(self.filename, str):
            pheno_input = pstpheno.loadPhen(self.filename, missing=self.missing)
        elif self.filename is None:
            assert (
                self._iid_if_none is not None
            ), "If input is None then iid_if_none be given"
            pheno_input = {
                "header": np.empty((0), dtype="str"),
                "vals": np.empty((len(self._iid_if_none), 0)),
                "iid": self._iid_if_none,
            }
        else:
            pheno_input = self.filename

        if len(pheno_input["vals"].shape) == 1:
            pheno_input = {
                "header": pheno_input["header"],
                "vals": np.reshape(pheno_input["vals"], (-1, 1)),
                "iid": pheno_input["iid"],
            }

        if len(pheno_input["header"]) > 0 and pheno_input["header"][0] is None:
            pheno_input["header"] = [
                "pheno{0}".format(i) for i in range(len(pheno_input["header"]))
            ]  # LATER move to reader?
        elif len(pheno_input["header"]) == 0:
            pheno_input["header"] = [
                "pheno{0}".format(i) for i in range(pheno_input["vals"].shape[1])
            ]

        row = pheno_input["iid"]
        col = np.array(pheno_input["header"], dtype="str")
        col_property = np.empty((len(col), 3))
        col_property.fill(np.nan)
        val = pheno_input["vals"]

        snpdata = SnpData(iid=row, sid=col, pos=col_property, val=val)
        return snpdata

    @staticmethod
    def write(filename, snpdata, missing="NaN", sep="\t"):
        """Writes a :class:`SnpData` to Pheno format and returns the :class:`.Pheno`.

        :param filename: the name of the file to create
        :type filename: string
        :param missing: value to threat as missing (default 'NaN')
        :type missing: string
        :param sep: the separator (default '\t')
        :type sep: string
        :rtype: :class:`.Pheno`


        >>> from pysnptools.snpreader import Pheno, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:,:10].read()  # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.phe")
        >>> Pheno.write("tempdir/toydata10.txt",snpdata)       # Write data in Pheno format
        Pheno('tempdir/toydata10.txt')
        """
        with open(filename, "w") as f:
            for i in range(snpdata.iid_count):
                tmpstr = snpdata.iid[i, 0] + sep + snpdata.iid[i, 1]
                for m in range(snpdata.sid_count):
                    v = snpdata.val[i, m]
                    if np.isnan(v):
                        vs = missing
                    else:
                        vs = str(v)
                    tmpstr += sep + vs
                tmpstr += "\n"
                f.write(tmpstr)
        return Pheno(filename, missing=missing)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
from itertools import *
import logging
from bed_reader import open_bed, to_bed
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData
import warnings


plink_chrom_map = {"X": 23, "Y": 24, "XY": 25, "MT": 26}

reverse_plink_chrom_map = {23: "X", 24: "Y", 25: "XY", 26: "MT"}


class Bed(SnpReader):
    """
    A :class:`.SnpReader` for random-access reads of Bed/Bim/Fam files from disk.

    See :class:`.SnpReader` for details and examples.

    The format is described `here <http://zzz.bwh.harvard.edu/plink/binary.shtml>`__.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The `*.bed` file to read. The '.bed' suffix is optional. The related `*.bim` and `*.fam` files will also be read.
                     * **count_A1** (*bool*) -- Tells if it should count the number of A1 alleles (the PLINK standard) or the number of A2 alleles. False is the current default, but in the future the default will change to True.

                     *The following three options are never needed, but can be used to avoid reading large '.fam' and '.bim' files when their information is already known.*

                     * **iid** (an array of strings) -- The :attr:`SnpReader.iid` information. If not given, reads info from '.fam' file.
                     * **sid** (an array of strings) -- The :attr:`SnpReader.sid` information. If not given, reads info from '.bim' file.
                     * **pos** (optional, an array of strings) -- The :attr:`SnpReader.pos` information.  If not given, reads info from '.bim' file.
                     * **num_threads** (optinal, int) -- The number of threads with which to read data. Defaults to all available processors.
                            Can also be set with these environment variables (listed in priority order):
                            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
                     * **skip_format_check** (*bool*) -- By default (False), checks that the '.bed' file starts with the expected bytes
                            the first time any file ('.bed', '.fam', or '.bim') is opened.
                     * **fam_filename** (optional, *string*) -- The `*.fam` file to read. Defaults to the bed filename with the suffix replaced.
                     * **bim_filename** (optional, *string*) -- The `*.bim` file to read. Defaults to the bed filename with the suffix replaced.
                     * **chrom_map** (optional, *dictionary*) -- A dictionary from non-numeric chromosome names to numbers. Defaults to the PLINK
                            mapping, namely, :data:`plink_chrom_map`.

    **Constants**

    * plink_chrom_map
        A dictionary that 'X' to 23, 'Y' to 24, 'XY' to 25, and 'MT' to 26.

    * reverse_plink_chrom_map
        A dictionary that maps 23 to 'X', 24 to 'Y', 25 to 'XY', and 26 to 'MT'

    **Methods beyond** :class:`.SnpReader`

        The :meth:`.SnpReader.read` method returns a :class:`SnpData` with a :attr:`SnpData.val` ndarray. By default, this ndarray is
        numpy.float32. Optionally, it can be numpy.float16. For :class:`Bed`, however, it can also be numpy.int8 with missing values
        represented by -127.

        When reading, any chromosome and position values of 0 (the PLINK standard for missing) will be represented in :attr:`pos` as NaN.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/distributed_bed_test1_X.*","*.bed")
        >>> snp_on_disk = Bed(bedfile, count_A1=False)
        >>> snpdata1 = snp_on_disk.read() # Read into the default, an float64 ndarray
        >>> snpdata1.val.dtype
        dtype('float64')
        >>> snpdata1 = snp_on_disk.read(dtype='int8',_require_float32_64=False) #Read into an 'int8' ndarray.
        >>> snpdata1.val.dtype
        dtype('int8')

    """

    def __init__(
        self,
        filename,
        count_A1=None,
        iid=None,
        sid=None,
        pos=None,
        num_threads=None,
        skip_format_check=False,
        fam_filename=None,
        bim_filename=None,
        chrom_map=plink_chrom_map,
    ):
        super(Bed, self).__init__()

        self._ran_once = False

        self.filename = SnpReader._name_of_other_file(
            filename, remove_suffix="bed", add_suffix="bed"
        )
        self.fam_filename = fam_filename
        self.bim_filename = bim_filename
        if count_A1 is None:
            warnings.warn(
                "'count_A1' was not set. For now it will default to 'False', but in the future it will default to 'True'",
                FutureWarning,
            )
            count_A1 = False
        self.count_A1 = count_A1
        self._skip_format_check = skip_format_check
        self._original_iid = iid
        self._original_sid = sid
        self._original_pos = pos
        self._num_threads = num_threads
        self._open_bed = None
        self.chrom_map = chrom_map

    def __repr__(self):
        return "{0}('{1}',count_A1={2})".format(
            self.__class__.__name__, self.filename, self.count_A1
        )

    def _open_bed_if_needed(self):
        if self._open_bed is not None:
            return

        properties = {
            "father": None,
            "mother": None,
            "sex": None,
            "pheno": None,
            "allele_1": None,
            "allele_2": None,
        }
        if self._original_iid is not None:
            properties["fid"] = self._original_iid[:, 0]
            properties["iid"] = self._original_iid[:, 1]
        if self._original_sid is not None:
            properties["sid"] = self._original_sid
        if self._original_pos is not None:
            properties["chromosome"] = self._original_pos[:, 0]
            properties["cm_position"] = self._original_pos[:, 1]
            properties["bp_position"] = self._original_pos[:, 2]

        self._open_bed = open_bed(
            self.filename,
            properties=properties,
            skip_format_check=self._skip_format_check,
            count_A1=self.count_A1,
            num_threads=self._num_threads,
            fam_filepath=self.fam_filename,
            bim_filepath=self.bim_filename,
        )

    @property
    def row(self):
        """*same as* :attr:`iid`"""
        if not hasattr(self, "_row"):
            self._open_bed_if_needed()

            self._row = np.array(
                [self._open_bed.fid, self._open_bed.iid]
            ).T  # LATER: could copy in batches or use concatenate

        return self._row

    @property
    def col(self):
        """*same as* :attr:`sid`"""
        if not hasattr(self, "_col"):
            self._open_bed_if_needed()
            self._col = self._open_bed.sid
        return self._col

    @property
    def col_property(self):
        """*same as* :attr:`pos`"""
        if not hasattr(self, "_col_property"):
            self._open_bed_if_needed()

            chromosome = self._open_bed.chromosome
            intersection = self.chrom_map.keys() & chromosome
            for key in intersection:
                chromosome[chromosome == key] = self.chrom_map[key]

            self._col_property = np.array(
                [
                    self._open_bed.chromosome.astype("float"),
                    self._open_bed.cm_position,
                    self._open_bed.bp_position,
                ]
            ).T  # LATER: Could copy in batches to use less memory
            self._col_property[self._col_property == 0] = np.nan

        return self._col_property

    def _run_once(self):
        if self._ran_once:
            return
        self._ran_once = True

        self.row
        self.col
        self.col_property
        assert self._open_bed is not None  # real assert

        self._assert_iid_sid_pos(check_val=False)

    def __del__(self):
        pass

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because only uses original inputs
        copier.input(
            SnpReader._name_of_other_file(
                self.filename, remove_suffix="bed", add_suffix="bed"
            )
        )
        copier.input(
            SnpReader._name_of_other_file(
                self.filename, remove_suffix="bed", add_suffix="bim"
            )
        )
        copier.input(
            SnpReader._name_of_other_file(
                self.filename, remove_suffix="bed", add_suffix="fam"
            )
        )

    @staticmethod
    def write(
        filename,
        snpdata,
        count_A1=False,
        force_python_only=False,
        _require_float32_64=True,
        num_threads=None,
        reverse_chrom_map={},
    ):
        """Writes a :class:`SnpData` to Bed format and returns the :class:`.Bed`.

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :param count_A1: Tells if it should count the number of A1 alleles (the PLINK standard) or the number of A2 alleles. False is the current default, but in the future the default will change to True.
        :type count_A1: bool
        :param force_python_only: Defaults to False. Tells to use Python code rather than faster Rust code. (Might be useful for debugging).
        :type force_python_only: bool
        :param _require_float32_64: Defaults to True. Requires that snpdata's dtype is float32 or float64. (False is useful for writing int8 data.)
        :type _require_float32_64: bool
        :param num_threads: Maximum number of threads to use. Currently ignored and number of threads is 1.
        :type num_threads: int
        :param reverse_chrom_map: Dictionary from chromosome number to chromsome string to write in the `*.bim` file. Defaults to empty dictionary.
        :type reverse_chrom_map: dictionary
        :rtype: :class:`.Bed`

        Any :attr:`pos` values of NaN will be written as 0, the PLINK standard for missing chromosome and position values.

        >>> from pysnptools.snpreader import Pheno, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_fn = example_file("pysnptools/examples/toydata.5chrom.*", "*.bed")
        >>> snpdata = Bed(bed_fn)[:,::2].read() # Read every-other SNP
        >>> pstutil.create_directory_if_necessary("tempdir/everyother.bed")
        >>> Bed.write("tempdir/everyother.bed",snpdata,count_A1=False)   # Write data in Bed format
        Bed('tempdir/everyother.bed',count_A1=False)
        >>> # Can write from an int8 array, too.
        >>> snpdata_int = SnpData(val=np.int_(snpdata.val).astype('int8'),iid=snpdata.iid,sid=snpdata.sid,pos=snpdata.pos,_require_float32_64=False)
        >>> snpdata_int.val.dtype
        dtype('int8')
        >>> Bed.write("tempdir/everyother.bed",snpdata_int,count_A1=False,_require_float32_64=False)
        Bed('tempdir/everyother.bed',count_A1=False)
        """

        if isinstance(filename, SnpData) and isinstance(
            snpdata, str
        ):  # For backwards compatibility, reverse inputs if necessary
            warnings.warn(
                "write statement should have filename before data to write",
                DeprecationWarning,
            )
            filename, snpdata = snpdata, filename

        if count_A1 is None:
            warnings.warn(
                "'count_A1' was not set. For now it will default to 'False', but in the future it will default to 'True'",
                FutureWarning,
            )
            count_A1 = False

        filename = SnpReader._name_of_other_file(
            filename, remove_suffix="bed", add_suffix="bed"
        )

        chromosome = snpdata.pos[:, 0]
        intersection = reverse_chrom_map.keys() & chromosome
        if len(intersection) > 0:
            chromosome = chromosome.astype("object")
            for key in intersection:
                chromosome[chromosome == key] = reverse_chrom_map[key]

        to_bed(
            filename,
            val=snpdata.val,
            properties={
                "fid": snpdata.iid[:, 0],
                "iid": snpdata.iid[:, 1],
                "sid": snpdata.sid,
                "chromosome": chromosome.copy(),
                "cm_position": snpdata.pos[:, 1].copy(),
                "bp_position": snpdata.pos[:, 2].copy(),
            },
            count_A1=count_A1,
            force_python_only=force_python_only,
            num_threads=num_threads,
        )

        return Bed(filename, count_A1=count_A1)

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()

        if order == "A":
            order = "F"

        assert not hasattr(
            self, "ind_used"
        ), "A SnpReader should not have a 'ind_used' attribute"

        val = self._open_bed.read(
            index=(iid_index_or_none, sid_index_or_none),
            order=order,
            dtype=dtype,
            force_python_only=force_python_only,
            num_threads=num_threads,
        )

        return val


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    import os

    if False:  # Look for example Bed files with missing data
        from pysnptools.util._example_file import pysnptools_hashdown
        from pysnptools.util import example_file

        for file in pysnptools_hashdown.walk():
            if file.endswith(".bed"):
                print(file + "?")
                bed_file = None
                try:
                    bed_file = example_file(file[:-4] + ".*", "*.bed")
                except Exception:
                    pass
                if bed_file is not None:
                    bed = Bed(bed_file)
                    snpdata = bed[:1000, :1000].read()
                    if not np.all(snpdata.val == snpdata.val):
                        print(bed_file + "!")

    if False:
        from pysnptools.snpreader import Bed
        from pysnptools.util import example_file  # Download and return local file name

        # bed_file = example_file('doc/ipynb/all.*','*.bed')
        bed_file = r"F:\backup\carlk4d\data\carlk\cachebio\genetics\onemil\id1000000.sid_1000000.seed0.byiid\iid990000to1000000.bed"
        bed = Bed(bed_file, count_A1=False)
        snpdata1 = bed[:, :1000].read()
        snpdata2 = bed[:, :1000].read(dtype="int8", _require_float32_64=False)
        print(snpdata2)
        snpdata3 = bed[:, :1000].read(
            dtype="int8", order="C", _require_float32_64=False
        )
        print(snpdata3)
        snpdata3.val = snpdata3.val.astype("float32")
        snpdata3.val.dtype

    if False:
        from pysnptools.snpreader import Bed, SnpGen

        iid_count = 487409
        sid_count = 5000
        sid_count_max = 5765294
        sid_batch_size = 50

        sid_batch_count = -(sid_count // -sid_batch_size)
        sid_batch_count_max = -(sid_count_max // -sid_batch_size)
        snpgen = SnpGen(seed=234, iid_count=iid_count, sid_count=sid_count_max)

        for batch_index in range(sid_batch_count):
            sid_index_start = batch_index * sid_batch_size
            sid_index_end = (batch_index + 1) * sid_batch_size  # what about rounding
            filename = r"d:\deldir\rand\fakeukC{0}x{1}-{2}.bed".format(
                iid_count, sid_index_start, sid_index_end
            )
            if not os.path.exists(filename):
                Bed.write(
                    filename + ".temp", snpgen[:, sid_index_start:sid_index_end].read()
                )
                os.rename(filename + ".temp", filename)

    if False:
        from pysnptools.snpreader import Pheno, Bed

        filename = r"m:\deldir\New folder (4)\all_chr.maf0.001.N300.bed"
        iid_count = 300
        iid = [["0", "iid_{0}".format(iid_index)] for iid_index in range(iid_count)]
        bed = Bed(filename, iid=iid, count_A1=False)
        print(bed.iid_count)

    if False:
        from pysnptools.util import example_file

        pheno_fn = example_file("pysnptools/examples/toydata.phe")

    if False:
        from pysnptools.snpreader import Bed
        import pysnptools.util as pstutil
        import os

        print(os.getcwd())
        snpdata = Pheno("../examples/toydata.phe").read()  # Read data from Pheno format
        pstutil.create_directory_if_necessary("tempdir/toydata.5chrom.bed")
        Bed.write(
            "tempdir/toydata.5chrom.bed", snpdata, count_A1=False
        )  # Write data in Bed format

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.distreader import DistReader
from pysnptools.standardizer import Unit
from pysnptools.standardizer import Identity
from pysnptools.pstreader import PstData
import warnings
import time


class DistData(PstData, DistReader):
    """
    A :class:`.DistReader` for holding SNP distributions (or similar values) in-memory, along with related *iid*, *sid*, and *pos* information.
    It is usually created by calling the :meth:`.DistReader.read` method on another :class:`.DistReader`, for example, :class:`.Bgen`.
    It can also be constructed directly.

    See :class:`.DistReader` for details and examples.

    **Constructor:**
        :Parameters: * **iid** (an array of string pair) -- The :attr:`DistReader.iid` information.
                     * **sid** (an array of strings) -- The :attr:`DistReader.sid` information.
                     * **val** (a 3-D array of floats) -- The SNP value distribution
                     * **pos** (optional, an array of strings) -- The :attr:`DistReader.pos` information
                     * **name** (optional, string) -- Information to be display about the origin of this data
                     * **copyinputs_function** (optional, function) -- *Used internally by optional clustering code*

        :Example:

        >>> from pysnptools.distreader import DistData
        >>> distdata = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[.5,.5,0]]])
        >>> print((distdata.val[1,1], distdata.iid_count, distdata.sid_count))
        (array([0.  , 0.75, 0.25]), 2, 3)

    **Equality:**

        Two DistData objects are equal if their four arrays (:attr:`DistData.val`, :attr:`DistReader.iid`, :attr:`DistReader.sid`, and :attr:`DistReader.pos`)
        are 'array_equal'. (Their 'name' does not need to be the same).
        If either :attr:`DistData.val` contains NaN, the objects will not be equal. However, :meth:`.DistData.allclose` can be used to treat NaN as
        regular values.

        :Example:

        >>> import numpy as np
        >>> from pysnptools.distreader import DistData
        >>> snpdata1 = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[.5,.5,0]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata2 = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[.5,.5,0]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata1 == snpdata2) #True, because all the arrays have the same values.
        True
        >>> print(snpdata1.val is snpdata2.val) #False, because the two arrays have different memory.
        False
        >>> snpdata3 = DistData(iid=[['a','0'],['b','0']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[.5,.5,0]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata4 = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[.5,.5,0]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata3 == snpdata4) #False, because the iids are different.
        False
        >>> snpdata5 = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[np.nan,np.nan,np.nan]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata6 = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[np.nan,np.nan,np.nan]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata5 == snpdata6) #False, because the val's contain NaN
        False
        >>> print(snpdata5.allclose(snpdata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True

    **Methods beyond** :class:`.DistReader`

    """

    def __init__(self, iid, sid, val, pos=None, name=None, copyinputs_function=None):
        # We don't have a 'super(DistData, self).__init__()' here because DistData takes full responsibility for initializing both its superclasses

        self._val = None

        self._row = PstData._fixup_input(
            iid, empty_creator=lambda ignore: np.empty([0, 2], dtype="str"), dtype="str"
        )
        self._col = PstData._fixup_input(
            sid, empty_creator=lambda ignore: np.empty([0], dtype="str"), dtype="str"
        )
        self._row_property = PstData._fixup_input(
            None,
            count=len(self._row),
            empty_creator=lambda count: np.empty([count, 0], dtype="str"),
            dtype="str",
        )
        self._col_property = PstData._fixup_input(
            pos,
            count=len(self._col),
            empty_creator=lambda count: np.full([count, 3], np.nan),
        )
        self._val = PstData._fixup_input_val(
            val,
            row_count=len(self._row),
            col_count=len(self._col),
            empty_creator=lambda row_count, col_count: np.empty(
                [row_count, col_count, 3], dtype=np.float64
            ),
        )  # !!!Replace empty with my FillNA method?
        self._assert_iid_sid_pos(check_val=True)
        self._name = name or ""
        self._std_string_list = []

    @property
    def val(self):
        """The 3D NumPy array of floats that represents the distribution of SNP values. You can get or set this property.

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> distdata = Bgen(bgen_file)[:5,:].read() #read data for first 5 iids
        >>> print(distdata.val[4,55]) #print one of the SNP values
        [0.23137255 0.65342184 0.11520562]
        """
        return self._val

    @val.setter
    def val(self, new_value):
        self._val = PstData._fixup_input_val(
            new_value,
            row_count=len(self._row),
            col_count=len(self._col),
            empty_creator=lambda row_count, col_count: np.empty(
                [row_count, col_count, 3], dtype=np.float64
            ),
        )  # !!!Replace empty with my FillNA method?
        self._assert_iid_sid_pos(check_val=True)

    def allclose(self, value, equal_nan=True):
        """
        :param value: Other object with which to compare.
        :type value: :class:`DistData`
        :param equal_nan: (Default: True) Tells if NaN in :attr:`DistData.val` should be treated as regular values when testing equality.
        :type equal_nan: bool

        >>> import numpy as np
        >>> snpdata5 = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[np.nan,np.nan,np.nan]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata6 = DistData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],
        ...                     val=[[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                          [[0,1.,0],[0,.75,.25],[np.nan,np.nan,np.nan]]],
        ...                     pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata5.allclose(snpdata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True
        >>> print(snpdata5.allclose(snpdata6,equal_nan=False)) #False, if we consider the NaN as special values, all the arrays are not equal.
        False

        """
        return PstData.allclose(self, value, equal_nan=equal_nan)

    def __repr__(self):
        if self._name == "":
            if len(self._std_string_list) > 0:
                s = "{0}({1})".format(
                    self.__class__.__name__, ",".join(self._std_string_list)
                )
            else:
                s = "{0}()".format(self.__class__.__name__)
        else:
            if len(self._std_string_list) > 0:
                s = "{0}({1},{2})".format(
                    self.__class__.__name__, self._name, ",".join(self._std_string_list)
                )
            else:
                s = "{0}({1})".format(self.__class__.__name__, self._name)
        return s


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod()
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import logging
import os
import shutil
import numpy as np
import unittest
import doctest
from pathlib import Path
import pysnptools.util as pstutil
from pysnptools.pstreader import PstData
from pysnptools.pstreader import PstMemMap
from pysnptools.snpreader import SnpReader, SnpData
from pysnptools.standardizer import Identity
from pysnptools.util import log_in_place


class SnpMemMap(PstMemMap, SnpData):
    """
    A :class:`.SnpData` that keeps its data in a memory-mapped file. This allows data large than fits in main memory.

    See :class:`.SnpData` for general examples of using SnpData.

    **Constructor:**
        :Parameters: **filename** (*string*) -- The *\*.snp.memmap* file to read.

        Also see :meth:`.SnpMemMap.empty` and :meth:`.SnpMemMap.write`.

        :Example:

        >>> from pysnptools.snpreader import SnpMemMap
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> mem_map_file = example_file('pysnptools/examples/tiny.snp.memmap')
        >>> snp_mem_map = SnpMemMap(mem_map_file)
        >>> print(snp_mem_map.val[0,1], snp_mem_map.iid_count, snp_mem_map.sid_count)
        2.0 2 3

    **Methods inherited from** :class:`.SnpData`

        :meth:`.SnpData.allclose`, :meth:`.SnpData.standardize`

    **Methods beyond** :class:`.SnpReader`

    """

    def __init__(self, *args, **kwargs):
        super(SnpMemMap, self).__init__(*args, **kwargs)

    @property
    def val(self):
        """The 2D NumPy memmap array of floats that represents the values. You can get this property, but cannot set it (except with itself)

        >>> from pysnptools.snpreader import SnpMemMap
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> mem_map_file = example_file('pysnptools/examples/tiny.snp.memmap')
        >>> snp_mem_map = SnpMemMap(mem_map_file)
        >>> print(snp_mem_map.val[0,1])
        2.0
        """
        self._run_once()
        return self._val

    @val.setter
    def val(self, new_value):
        self._run_once()
        if self._val is new_value:
            return
        raise Exception("SnpMemMap val's cannot be set to a different array")

    @property
    def offset(self):
        """The byte position in the file where the memory-mapped values start.

        (The disk space before this is used to store :attr:`SnpReader.iid`, etc. information.
        This property is useful when interfacing with, for example, external Fortran and C matrix libraries.)

        """
        self._run_once()
        return self._offset

    @property
    def filename(self):
        """The name of the memory-mapped file"""
        # Don't need '_run_once'
        return self._filename

    @staticmethod
    def empty(iid, sid, filename, pos=None, order="F", dtype=np.float64):
        """Create an empty :class:`.SnpMemMap` on disk.

        :param iid: The :attr:`SnpReader.iid` information
        :type iid: an array of string pairs

        :param sid: The :attr:`SnpReader.sid` information
        :type sid: an array of strings

        :param filename: name of memory-mapped file to create
        :type filename: string

        :param pos: optional -- The additional :attr:`SnpReader.pos` information associated with each sid. Default: None
        :type pos: an array of numeric triples

        :param order: {'F' (default), 'C'}, optional -- Specify the order of the ndarray.
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`SnpMemMap.val` ndarray.
        :type dtype: data-type

        :rtype: :class:`.SnpMemMap`

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.snpreader import SnpMemMap
        >>> filename = "tempdir/tiny.snp.memmap"
        >>> pstutil.create_directory_if_necessary(filename)
        >>> snp_mem_map = SnpMemMap.empty(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],filename=filename,order="F",dtype=np.float64)
        >>> snp_mem_map.val[:,:] = [[0.,2.,0.],[0.,1.,2.]]
        >>> snp_mem_map.flush()

        """

        self = SnpMemMap(filename)
        self._empty_inner(
            row=iid,
            col=sid,
            filename=filename,
            row_property=None,
            col_property=pos,
            order=order,
            dtype=dtype,
            val_shape=None,
        )
        return self

    def flush(self):
        """Flush :attr:`SnpMemMap.val` to disk and close the file. (If values or properties are accessed again, the file will be reopened.)

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.snpreader import SnpMemMap
        >>> filename = "tempdir/tiny.snp.memmap"
        >>> pstutil.create_directory_if_necessary(filename)
        >>> snp_mem_map = SnpMemMap.empty(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],filename=filename,order="F",dtype=np.float64)
        >>> snp_mem_map.val[:,:] = [[0.,2.,0.],[0.,1.,2.]]
        >>> snp_mem_map.flush()

        """
        if self._ran_once:
            self.val.flush()
            del self._val
            self._ran_once = False

    @staticmethod
    def write(
        filename,
        snpreader,
        standardizer=Identity(),
        order="A",
        dtype=None,
        block_size=None,
        num_threads=None,
    ):
        """Writes a :class:`SnpReader` to :class:`SnpMemMap` format.

        :param filename: the name of the file to create
        :type filename: string
        :param snpreader: The data that should be written to disk.
        :type snpreader: :class:`SnpReader`
        :rtype: :class:`.SnpMemMap`

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> from pysnptools.snpreader import Bed, SnpMemMap
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> bed = Bed(bed_file)
        >>> pstutil.create_directory_if_necessary("tempdir/toydata.5chrom.snp.memmap") #LATER should we just promise to create directories?
        >>> SnpMemMap.write("tempdir/toydata.5chrom.snp.memmap",bed)      # Write bed in SnpMemMap format
        SnpMemMap('tempdir/toydata.5chrom.snp.memmap')
        """
        block_size = block_size or max((100_000) // max(1, snpreader.row_count), 1)

        if hasattr(snpreader, "val"):
            order = PstMemMap._order(snpreader) if order == "A" else order
            dtype = dtype or snpreader.val.dtype
        else:
            order = "F" if order == "A" else order
            dtype = dtype or np.float64
        dtype = np.dtype(dtype)

        snpmemmap = SnpMemMap.empty(
            iid=snpreader.iid,
            sid=snpreader.sid,
            filename=filename + ".temp",
            pos=snpreader.col_property,
            order=order,
            dtype=dtype,
        )
        if hasattr(snpreader, "val"):
            standardizer.standardize(snpreader, num_threads=num_threads)
            snpmemmap.val[:, :] = snpreader.val
        else:  # !!
            with log_in_place("SnpMemMap write sid# !!dex ", logging.INFO) as updater:
                for start in range(0, snpreader.sid_count, block_size):
                    updater("{0} of {1}".format(start, snpreader.sid_count))
                    snpdata = snpreader[:, start : start + block_size].read(
                        order=order, dtype=dtype, num_threads=num_threads
                    )
                    standardizer.standardize(snpdata, num_threads=num_threads)
                    snpmemmap.val[:, start : start + snpdata.sid_count] = snpdata.val

        snpmemmap.flush()
        if os.path.exists(filename):
            os.remove(filename)
        shutil.move(filename + ".temp", filename)
        logging.debug("Done writing " + filename)
        return SnpMemMap(filename)

    def _run_once(self):
        if self._ran_once:
            return
        row_ascii, col_ascii, val, row_property, col_property = self._run_once_inner()
        row = np.array(row_ascii, dtype="str")  # !!!avoid this copy when not needed
        col = np.array(col_ascii, dtype="str")  # !!!avoid this copy when not needed

        SnpData.__init__(
            self,
            iid=row,
            sid=col,
            val=val,
            pos=col_property,
            name="np.memmap('{0}')".format(self._filename),
        )


class TestSnpMemMap(unittest.TestCase):
    def test1(self):
        from pysnptools.snpreader import Bed, SnpMemMap
        from pysnptools.util import example_file  # Download and return local file name

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        filename2 = "tempdir/tiny.snp.memmap"
        pstutil.create_directory_if_necessary(filename2)
        snpreader2 = SnpMemMap.empty(
            iid=[["fam0", "iid0"], ["fam0", "iid1"]],
            sid=["snp334", "snp349", "snp921"],
            filename=filename2,
            order="F",
            dtype=np.float64,
        )
        assert isinstance(snpreader2.val, np.memmap)
        snpreader2.val[:, :] = [[0.0, 2.0, 0.0], [0.0, 1.0, 2.0]]
        assert np.array_equal(
            snpreader2[[1], [1]].read(view_ok=True).val, np.array([[1.0]])
        )
        snpreader2.flush()
        assert isinstance(snpreader2.val, np.memmap)
        assert np.array_equal(
            snpreader2[[1], [1]].read(view_ok=True).val, np.array([[1.0]])
        )
        snpreader2.flush()

        snpreader3 = SnpMemMap(filename2)
        assert np.array_equal(
            snpreader3[[1], [1]].read(view_ok=True).val, np.array([[1.0]])
        )
        assert isinstance(snpreader3.val, np.memmap)

        logging.info("in TestSnpMemMap test1")
        snpreader = SnpMemMap("tempdir/tiny.snp.memmap")
        assert snpreader.iid_count == 2
        assert snpreader.sid_count == 3
        assert isinstance(snpreader.val, np.memmap)

        snpdata = snpreader.read(view_ok=True)
        assert isinstance(snpdata.val, np.memmap)

        bed_file = example_file("pysnptools/examples/toydata.5chrom.*", "*.bed")
        bed = Bed(bed_file)
        pstutil.create_directory_if_necessary(
            "tempdir/toydata.5chrom.snp.memmap"
        )  # LATER should we just promise to create directories?
        SnpMemMap.write(
            "tempdir/toydata.5chrom.snp.memmap", bed
        )  # Write bed in SnpMemMap format
        SnpMemMap.write(
            "tempdir/toydata.5chromsnpdata.snp.memmap", bed[:, ::2].read()
        )  # Write snpdata in SnpMemMap format

        os.chdir(old_dir)


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSnpMemMap))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARN)

    if False:
        from pysnptools.snpreader import Bed, _MergeSIDs, SnpMemMap

        # Data will appear in the memory mapped file in the order given.
        # The *.fam and *.bim files must be in the same order as the bed files.
        bed_file_list = []
        fam_file_list = []
        bim_file_list = []
        for piece in range(25):
            bed_file_list += [
                r"M:\deldir\testsnps_1_10_250000_10000\chrom10.piece{0}of25.bed".format(
                    piece
                )
            ]
            fam_file_list += [
                r"M:\deldir\testsnps_1_10_250000_10000\chrom10.piece{0}of25.fam".format(
                    piece
                )
            ]
            bim_file_list += [
                r"M:\deldir\testsnps_1_10_250000_10000\chrom10.piece{0}of25.bim".format(
                    piece
                )
            ]
        # for chrom in range(21,23):
        #    bed_file_list += [r"d:\deldir\genbgen\merged_487400x220000.{0}.bed".format(chrom)]
        #    fam_file_list += [r"m:\deldir\genbgen\merged_487400x220000.{0}.fam".format(chrom)]
        #    bim_file_list += [r"m:\deldir\genbgen\merged_487400x220000.{0}.bim".format(chrom)]

        memmap_file = r"D:\deldir\memmap1.snp.memmap"

        #######
        # For this demo, erase the memmap output file
        ######

        if Path(memmap_file).exists():
            Path(memmap_file).unlink()

        #######
        # Merge the input files
        ######
        merge = _MergeSIDs(
            [
                Bed(
                    bed_file,
                    fam_filename=fam_file,
                    bim_filename=bim_file,
                    count_A1=True,
                    skip_format_check=True,
                )
                for bed_file, fam_file, bim_file in zip(
                    bed_file_list, fam_file_list, bim_file_list
                )
            ]
        )

        # memmap = _bed_to_memmap2(merge,memmap_file=memmap_file,dtype='float32',step=10)
        from pysnptools.standardizer import Unit

        memmap = SnpMemMap.write(
            memmap_file, merge, standardizer=Unit(), dtype="float32"
        )
        memmap

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    result = doctest.testmod(optionflags=doctest.ELLIPSIS)
    assert result.failed == 0, "failed doc test: " + __file__



import numpy as np
import logging
import pysnptools.util as pstutil
from bed_reader import get_num_threads, standardize_f64, standardize_f32


class Standardizer(object):
    """
    A Standardizer is a class such as :class:`.Unit` and :class:`.Beta` to be used by the :meth:`.SnpData.standardize` and :meth:`.SnpReader.read_kernel` method to standardize SNP data.

    :Example:

    Read and standardize SNP data.

    >>> from pysnptools.standardizer import Unit
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> snpdata1 = Bed(bedfile,count_A1=False).read().standardize(Unit())
    >>> print('{0:.6f}'.format(snpdata1.val[0,0]))
    0.229416

    Create a kernel from SNP data on disk.

    >>> bedfile2 = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
    >>> kerneldata = Bed(bedfile2,count_A1=False).read_kernel(Unit())
    >>> print('{0:.6f}'.format(kerneldata.val[0,0]))
    9923.069928


    Can also return a constant SNP standardizer that can be applied to other :class:`.SnpData`.

    >>> snp_whole = Bed(bedfile,count_A1=False)
    >>> train_idx, test_idx = range(10,snp_whole.iid_count), range(0,10) #test on the first 10, train on the rest
    >>> snp_train, trained_standardizer = Unit().standardize(snp_whole[train_idx,:].read(),return_trained=True)
    >>> print('{0:.6f}'.format(snp_train.val[0,0]))
    0.233550
    >>> print(trained_standardizer.stats[0,:]) #The mean and stddev of the 1st SNP on the training data # '...' for a possible space character
    [...1.94827586  0.22146953]
    >>> snp_test = snp_whole[test_idx,:].read().standardize(trained_standardizer)
    >>> snp_test.val[0,0]
    0.23354968324845735

    Standardize any Numpy array.

    >>> val = Bed(bedfile,count_A1=False).read().val
    >>> print(val[0,0])
    2.0
    >>> val = Unit().standardize(val)
    >>> print('{0:.6f}'.format(val[0,0]))
    0.229416

    Details of Methods & Properties:
    """

    def __init__(self):
        super(Standardizer, self).__init__()

    def standardize(
        self,
        snps,
        block_size=None,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        """
        Applies standardization, in place, to :class:`.SnpData` (or a NumPy array). For convenience also returns the :class:`.SnpData` (or a NumPy array).

        :param snps: SNP values to standardize
        :type snps: :class:`.SnpData` (or a NumPy array)

        :param block_size: *Not used*
        :type block_size: None

        :param return_trained: If true, returns a second value containing a constant standardizer trained on this data.
        :type return_trained: bool

        :param force_python_only: optional -- If False (default), may use outside library code. If True, requests that the read
            be done without outside library code.
        :type force_python_only: bool

        :param num_threads: optional -- The number of threads with which to standardize data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int

        :rtype: :class:`.SnpData` (or a NumPy array), (optional) constant :class:`.Standardizer`

        """
        if block_size is not None:
            warnings.warn(
                "block_size is deprecated (and not needed, since standardization is in-place",
                DeprecationWarning,
            )
        raise NotImplementedError(
            "subclass {0} needs to implement method '.standardize'".format(
                self.__class__.__name__
            )
        )

    @staticmethod
    # changes snps in place
    def _standardize_unit_and_beta(
        snps,
        is_beta,
        a,
        b,
        apply_in_place,
        use_stats,
        stats,
        num_threads,
        force_python_only=False,
    ):
        """
        When snps is a cupy ndarray, will use cupy to compute new stats for unit. (Other paths are not defined for cupy)
        """
        xp = pstutil.get_array_module(snps)

        assert (
            snps.flags["C_CONTIGUOUS"] or snps.flags["F_CONTIGUOUS"]
        ), "Expect snps to be order 'C' or order 'F'"

        # Make sure stats is the same type as snps. Because we might be creating a new array, we return it
        if stats is None:
            stats = xp.empty(
                [snps.shape[1], 2],
                dtype=snps.dtype,
                order="F" if snps.flags["F_CONTIGUOUS"] else "C",
            )
        elif not (
            stats.dtype == snps.dtype  # stats must have the same dtype as snps
            and (stats.flags["OWNDATA"])  # stats must own its data
            and (snps.flags["C_CONTIGUOUS"] and stats.flags["C_CONTIGUOUS"])
            or (
                snps.flags["F_CONTIGUOUS"] and stats.flags["F_CONTIGUOUS"]
            )  # stats must have the same order as snps
        ):
            stats = xp.array(
                stats,
                dtype=snps.dtype,
                order="F" if snps.flags["F_CONTIGUOUS"] else "C",
            )
        assert stats.shape == (snps.shape[1], 2), "stats must have size [sid_count,2]"

        if not force_python_only and xp is np:
            num_threads = get_num_threads(num_threads)

            if snps.dtype == np.float64:
                if (snps.flags["F_CONTIGUOUS"] or snps.flags["C_CONTIGUOUS"]) and (
                    snps.flags["OWNDATA"] or snps.base.nbytes == snps.nbytes
                ):  # !!create a method called is_single_segment
                    standardize_f64(
                        snps,
                        is_beta,
                        a,
                        b,
                        apply_in_place,
                        use_stats,
                        stats,
                        num_threads,
                    )
                    return stats
                else:
                    logging.info(
                        "Array is not contiguous, so will standardize with python only instead of C++"
                    )
            elif snps.dtype == np.float32:
                if (snps.flags["F_CONTIGUOUS"] or snps.flags["C_CONTIGUOUS"]) and (
                    snps.flags["OWNDATA"] or snps.base.nbytes == snps.nbytes
                ):
                    standardize_f32(
                        snps,
                        is_beta,
                        a,
                        b,
                        apply_in_place,
                        use_stats,
                        stats,
                        num_threads,
                    )
                    return stats
                else:
                    logging.info(
                        "Array is not contiguous, so will standardize with python only instead of C++"
                    )
            else:
                logging.info(
                    "Array type is not float64 or float32, so will standardize with python only instead of C++"
                )

        import pysnptools.standardizer as stdizer

        if is_beta:
            Standardizer._standardize_beta_python(
                snps, a, b, apply_in_place, use_stats=use_stats, stats=stats
            )
            return stats
        else:
            Standardizer._standardize_unit_python(
                snps, apply_in_place, use_stats=use_stats, stats=stats
            )
            return stats

    @staticmethod
    def _standardize_unit_python(snps, apply_in_place, use_stats, stats):
        """
        Standardize snps to zero-mean and unit variance.

        Will work with both numpy and cupy ndarray.
        """
        assert snps.dtype in [
            np.float64,
            np.float32,
        ], "snps must be a float in order to standardize in place."
        xp = pstutil.get_array_module(snps)

        imissX = xp.isnan(snps)

        if use_stats:
            snp_mean = stats[:, 0]
            snp_std = stats[:, 1]
        else:
            snp_std = xp.nanstd(snps, axis=0)
            snp_mean = xp.nanmean(snps, axis=0)
            # avoid div by 0 when standardizing
            # Don't need this warning because SNCs are still meaning full in QQ plots because they should be thought of as SNPs without enough data.
            # logging.warn("A least one snps has only one value, that is, its standard deviation is zero")
            snp_std[
                snp_std == 0.0
            ] = (
                xp.inf
            )  # We make the stdev infinity so that applying as a trained_standardizer will turn any input to 0. Thus if a variable has no variation in the training data, then it will be set to 0 in test data, too.
            stats[:, 0] = snp_mean
            stats[:, 1] = snp_std

        if apply_in_place:
            snps -= snp_mean
            snps /= snp_std
            snps[imissX] = 0

    @property
    def is_constant(self):
        """
        Tell if a standardizer's statistics are constant. For example, :class:`.Unit_Trained` contains a pre-specified mean and stddev for each SNP, so :attr:`Unit_Trained.is_constant` is True.
        :class:`.Unit`, on the other hand, measures the mean and stddev for each SNP, so :attr:`Unit.is_constant` is False.

        :rtype: bool
        """
        return False

    @staticmethod
    def _standardize_beta_python(snps, betaA, betaB, apply_in_place, use_stats, stats):
        """
        standardize snps with Beta prior
        """
        assert snps.dtype in [
            np.float64,
            np.float32,
        ], "snps must be a float in order to standardize in place."

        imissX = np.isnan(snps)
        snp_sum = np.nansum(snps, axis=0)
        n_obs_sum = (~imissX).sum(0)

        if use_stats:
            snp_mean = stats[:, 0]
            snp_std = stats[:, 1]
        else:
            snp_mean = (snp_sum * 1.0) / n_obs_sum
            snp_std = np.sqrt(np.nansum((snps - snp_mean) ** 2, axis=0) / n_obs_sum)
            if 0.0 in snp_std:
                # Don't need this warning because SNCs are still meaning full in QQ plots because they should be thought of as SNPs without enough data.
                # logging.warn("A least one snps has only one value, that is, its standard deviation is zero")
                snp_std[snp_std == 0] = np.inf
            stats[:, 0] = snp_mean
            stats[:, 1] = snp_std

        if apply_in_place:
            maf = snp_mean / 2.0
            maf[maf > 0.5] = 1.0 - maf[maf > 0.5]

            # avoid div by 0 when standardizing
            import scipy.stats as st

            maf_beta = st.beta.pdf(maf, betaA, betaB)
            # print("BetaPdf[{0},{1},{2}]={3}".format(maf,betaA,betaB,maf_beta))
            snps -= snp_mean
            snps *= maf_beta
            snps[imissX] = 0.0
            if (
                use_stats
            ):  # If we're applying to test data, set any variables with to 0 if they have no variation in the training data.
                snps[:, snp_std == np.inf] = 0.0

    def _merge_trained(self, trained_list):
        raise Exception("Not defined")


class _CannotBeTrained(Standardizer):
    def __init__(self, name):
        super(_CannotBeTrained, self).__init__()
        self.name = name

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, self.name)

    def standardize(
        self,
        snps,
        block_size=None,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        if block_size is not None:
            warnings.warn(
                "block_size is deprecated (and not needed, since standardization is in-place",
                DeprecationWarning,
            )
        raise Exception("Standardizer '{0}' cannot be trained", self)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
